package edu.udo.cs.ie.cowbots.bdi.brf;

import java.util.ArrayList;
import java.util.List;
import edu.udo.cs.ie.cowbots.bdi.brf.*;
import edu.udo.cs.ie.cowbots.logic.*;
import jason.asSyntax.Literal;

/**
 * 
 * annotator class. the annotator processes all incoming information
 * (percepts or communication), transforms these information into
 * an extended logical program and adds meta information to that
 * program (the result is an information object) 
 * 
 * @author Thomas
 *
 */
public class Annotator {

	// holds current world time. it is used to create a coarse order
	// over all information objects.
	int worldtime;		
	
	// sub worldtime counter, this allows the annotator to give
	// order information at a finer level than the world time.
	int subTime;	
	
		
	/**
	 * default constructor 
	 */
	public Annotator() {
		worldtime = 0;
		subTime = 0;
	}
	
	/**
	 * this function returns the current world time
	 * 
	 * @return		world time
	 */
	public int getWorldTime() {
		return worldtime;
	}
	
	public int getAndIncreaseSubtime() {
		return subTime++;
	}

	/**
	 * this function creates an information object from a
	 * jason literal
	 * 
	 * @param 	l	jason literal
	 * @return		information object containing l
	 */
	public InformationObject createIO(Literal l) {		
		Meta m = new Meta(worldtime);
		return new InformationObject(l,m);
	}
	
	/**
	 * this function creates an information object from an
	 * elp literal
	 * 
	 * @param 	l	jason literal
	 * @return		information object containing l
	 */
	public InformationObject createIO(ELPLiteral l) {
		ELP elp = new ELP();
		ELPRule r = new ELPRule(1,0);
		r.addHead(l);
		elp.addRule(r);
		Meta m = new Meta(worldtime);
		return new InformationObject(elp,m);
	}
	
	/**
	 * refreshes the percepted world time
	 * @param wt	world time
	 */
	public void setWorldTime(int wt) {
		this.worldtime = wt;
		this.subTime = 0;
	}
}
