package edu.udo.cs.ie.cowbots.bdi.brf;

import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.util.*;
import java.util.*;
import jason.asSyntax.*;
import jason.asSemantics.*;
import edu.udo.cs.ie.cowbots.bdi.brf.*;

/**
 * basic interface for any class that represents
 * an agent's epistemic state.
 * 
 * @author Thomas, Fabian, Boguslaw.
 *
 */
public interface IEpistemicState {
	
	/**
	 * returns the beliefs (a list of Literals) representing
	 * an agent's belief set. the addRules parameter lets
	 * the caller specify additional inference rules.
	 * 
	 * @param addRules	(optional) inference rules
	 * @return			belief set
	 */
	public List<Literal> getBeliefs(ELP addRules);
	
	/**
	 * returns the belief set as an elp
	 */
	public ELP getBeliefsAsELP();
	
		
	/**
	 * this function returns the queue of unprocessed
	 * messages
	 */
	public Queue<Message>	getMessages();
	
	
	/**
	 * this function returns the beliefs about an
	 * agents environment as a gridmap.
	 * 
	 * @return	gridmap with beliefs about the environment
	 */
	public GridMap	getMap();
	
	
	/**
	 * this method returns an agents belief set.
	 * 
	 * @return belief set of an agent
	 */
	public BeliefSet getBeliefSet();
	
	/**
	 * this method returns a list of new, but not
	 * not necessarily revised, perceptions.
	 * 
	 * @return collection of recent observations
	 */
	public List<ELPLiteral> getExploredCells();
	
	
	/**
	 * returns the time of the world (this is the
	 * time the agent beliefs in)
	 * 
	 * @return	time of world
	 */
	public int getWorldTime();
	
		
	
	/**
	 * returns a user program identified by a
	 * unique name (user programs are not part
	 * of the belief revision process over
	 * the epistemic state)
	 *   
	 * @param name	unique name of program
	 * @return elp matching given name or null 
	 */
	public ELP getProgram(String name);
}
