package edu.udo.cs.ie.cowbots.bdi.brf;

import java.util.*;

import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.logic.solver.*;


/**
 * this interface models a basic inference operator. an
 * inference operator is responsible for creating a
 * belief state (a consistent elp), selecting an answer set 
 * from the elp's solutions, clear extended literals from
 * that answer set and provide a list of rejected literals
 * or rules.
 * 
 * @author Thomas
 *
 */
public interface IInferenceOperator {

	/**
	 * this method builds a belief state from
	 * a list of information objects.
	 * 
	 * @param ioList	list of information objects
	 * @return			belief state (encoded as an elp)
	 */
	public ELP createBeliefState(List<InformationObject> ioList);
		
	
	/**
	 * this method evaluates a collection of answer sets
	 * and returns the "best" answer set. it is used in
	 * situations where multiple answer sets arise from
	 * an asp solver. additionally, any literals introduced
	 * during belief state construction, which are part
	 * of the inference's extended alphabet has to be
	 * removed here!
	 * 
	 * @param asl a list of answer sets
	 * @param dlv instance to dlv
	 * @return	answer set to be transformed into a belief set
	 */
	public AnswerSet selectAnswerSet(List<AnswerSet> asl, DLV dlv);
	
	
	/**
	 * returns the literals being rejected in the answer set
	 * analyzed by selectAnswerSet.
	 * 
	 * @return	functor-hashed lists of rejected literals
	 */
	public Map<String, List<ELPLiteral>> getRejectSet();
	
	
	/**
	 * assigns a set of "save" functor names to the inference.
	 * a save predicate is assumed to cause no conflicts and
	 * can be written directly into the edb (and answer set).
	 * 
	 * @param savePreds set of predicates which should not be transformed
	 */
	public void setSavePredicates(Set<String> savePreds);
}
