package edu.udo.cs.ie.cowbots.bdi.brf;

import java.util.List;
import java.util.Set;

/**
 * this interface models an identifier
 * table, which is a lookup table for
 * first terms of predicates assigned
 * to operators implementing this
 * interface. the literals stored by
 * IdentTable operators are called
 * identifiers.
 * 
 * @author Thomas
 *
 */
public interface IdentTable {

	/**
	 * checks if a given term is true for
	 * a given identifier
	 * @param term	argument to check
	 * @param identPredicate	identifier predicate
	 * @return	true if identPredicate(term,...) holds
	 * 
	 */
	public boolean isKindOf(String term, String identPredicate);
	
	/**
	 * returns all known predicates associated
	 * with a term
	 * @param term	term to lookup
	 * @return	list of predicates like p(term,...)
	 */
	public List<String> allIdentifiers(String term);
	
	
	/**
	 * returns the set of functor names that represent
	 * all identifier predicates
	 * @return	set of identifier functors
	 */
	public Set<String> identPredicates();
}
