package edu.udo.cs.ie.cowbots.bdi.brf.operators;

import java.util.*;

import org.w3c.dom.Element;

import edu.udo.cs.ie.cowbots.bdi.brf.*;
import edu.udo.cs.ie.cowbots.logic.*;

/**
 * this class is the default implementation of 
 * an IdentTable operator. it associates identifiers
 * with the first term, and assures that each
 * unique identifier is only stored once in
 * the belief base.
 * 
 * @author Thomas
 *
 */
public class IdentTableOperator implements IKnowledgeOperator, IdentTable {

	Set<String>					identPredicates;
	Set<String>					identifiers;
	Map<String,List<String>>	termLookup;
	
	/**
	 * default constructor
	 */
	public IdentTableOperator() {
		identifiers = new HashSet<String>();
		termLookup = new HashMap<String,List<String>>();
		identPredicates = new HashSet<String>();
	}
	
	@Override
	public void init(Element config) {
		// operator does not support any configuration
	}

	@Override
	public boolean update(List<InformationObject> updates, BeliefBase bb) {

		List<InformationObject> ioList = bb.getOperatorIO(this);

		// for any predicate foo(BAR), store BAR as key
		// in the termLookup table, and foo in the key's
		// value list. additionally, add foo to the list
		// of identPredicates
		for (InformationObject io : updates) {
			ELP elp = io.getELP();
			ELPAtom atm = elp.getRule(0).getHead(0).getAtom();
			String ident = atm.toString();
			if (!identifiers.contains(ident)) {
				// add identifier
				ioList.add(io);
				identifiers.add(ident);
				
				// link predicate with that
				String key = atm.getTerm(0);
				String val = atm.getPredicate();
				List<String> lst = termLookup.get(key);
				if (lst == null) {
					lst = new LinkedList<String>();
					termLookup.put(key, lst);
				}
				lst.add(val);
				identPredicates.add(val);
			}
		}
		return false;
	}

	@Override
	public boolean isKindOf(String term, String identPredicate) {
		List<String> lst = termLookup.get(term);
		
		if (lst == null)
			return false;
		
		for (String s : lst) {
			if (s.equalsIgnoreCase(identPredicate))
				return true;
		}
		
		return false;
	}
	
	@Override
	public List<String> allIdentifiers(String term) {
		return termLookup.get(term);
	}

	/**
	 * the optional operation deleteFacts is not
	 * supported by this operator.
	 */
	@Override
	public void deleteFacts(Set<ELPLiteral> toDel) {
		// default ident table does not support
		// deletion of literals
	}

	@Override
	public Set<String> identPredicates() {
		return this.identPredicates;
	}

	@Override
	public void processBeliefSet(BeliefSet bs, BeliefBase base) {
		// default ident table does not do anything
	}

	@Override
	public ELP getStateProgram() {
		return null;
	}

	@Override
	public void reset() {
		identifiers = new HashSet<String>();
		termLookup = new HashMap<String,List<String>>();
		identPredicates = new HashSet<String>();
	}
}
