package edu.udo.cs.ie.cowbots.bdi.brf.operators;

import java.awt.Point;
import java.util.*;

import org.w3c.dom.Element;

import edu.udo.cs.ie.cowbots.bdi.brf.*;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.util.*;


/**
 * a specialized operator for handling map updates.
 * most parts of the map are empty or don't change
 * at all, like obstacles. an inference would reject
 * multiple those cell over and over, so it might
 * be better to handle these things at the belief
 * base level.  
 * 
 * @author Thomas
 *
 */
public class MapUpdate implements IKnowledgeOperator, MapOperator {

	GridMap						map;
	List<ELPLiteral>			newExplorations;
	List<InformationObject>		bels;
	Set<String> 				fenceCoords;
	Set<String>					switchCoords;
	
	public MapUpdate() {
		map = null;
		newExplorations = new LinkedList<ELPLiteral>();
		fenceCoords = new HashSet<String>();
		switchCoords = new HashSet<String>();
	}
		
	@Override
	public void init(Element config) {
	}

	@Override
	public boolean update(List<InformationObject> updates, BeliefBase bb) {
		
		int cellcount = 0;
		int objcount = 0;
		int explored = 0;
		
		IdentTable it = bb.getIdentifiers();
		List<ELPLiteral> oldMap = new LinkedList<ELPLiteral>();
		bels = bb.getOperatorIO(this);
		
		this.newExplorations.clear();
		this.fenceCoords.clear();
		
		for(InformationObject io : updates) {
			
			// old content
			ELPLiteral old = null;			
			
			// fetch first literal
			ELPLiteral obj = io.getELP().getRule(0).getHead(0);
			ELPAtom atm = obj.getAtom();
			String functor = atm.getPredicate();
			String objtype = atm.getTerm(0);
			
			if (functor.equalsIgnoreCase("ison")) {
				int x = Integer.parseInt(obj.getAtom().getTerm(1));
				int y = Integer.parseInt(obj.getAtom().getTerm(2));
							
				if ( it.isKindOf(objtype , "cow")) { 														
					// percepted a cow, always add as exploration
					this.newExplorations.add(obj);
					this.newExplorations.add(new ELPAtom("cow",objtype));
					bels.add(io);							
					++objcount;
					
					
				} else if (it.isKindOf(objtype , "agent")) {

					// percepted an agent
					bels.add(io);
					++objcount;
				} else {
					
					// some kind of obstacle, just put it on the map
					old = map.setContent(x, y, obj);
					
					++cellcount;
					
					// but allow switches and fences to be in sight
					if (it.isKindOf(objtype , "fence") ) /*
						|| it.isKindOf(objtype , "switch")
						|| it.isKindOf(objtype , "corral")) */ {
						this.fenceCoords.add( x+"_"+y );						
						bels.add(io);
					} else if (it.isKindOf(objtype, "switch")) {
						this.switchCoords.add( x+"_"+y );
						bels.add(io);
					} else	if (it.isKindOf(objtype, "corral")) {
						bels.add(io);
					}
					
					// send newly explored cells
					if (old != null) {
					//if (true) {	
						if ( it.isKindOf( old.getAtom().getTerm(0), "unexplored") ) {
						//if (true) {
							explored++;

							if (it.isKindOf(objtype ,"fence")) {
								this.newExplorations.add( obj );
								this.newExplorations.add( new ELPAtom("fence","fence") );						
								this.fenceCoords.add( obj.getAtom().getTerm(1) +"_" +obj.getAtom().getTerm(2) );
							}
							if (it.isKindOf(objtype , "obstacle")) {
								this.newExplorations.add( obj );
								this.newExplorations.add( new ELPAtom("obstacle","obstacle") );
							}
							if (it.isKindOf(objtype , "switch")) {
								this.newExplorations.add( obj );
								this.newExplorations.add( new ELPAtom("switch",objtype) );
							}
							if (it.isKindOf(objtype, "empty")) {
								this.newExplorations.add(obj);
								this.newExplorations.add( new ELPAtom("empty","empty") );
							}
						}
						
					}
				}					
			}
			
			if (functor.equalsIgnoreCase("isopen")) {
				int x = Integer.parseInt(obj.getAtom().getTerm(0));
				int y = Integer.parseInt(obj.getAtom().getTerm(1));
				
				bels.add(io);
				map.setFlag(x, y, obj);
				
				if (this.fenceCoords.contains( ""+x+"_"+y)) {
					this.newExplorations.add(obj);
				}
			}
		}
		
		return false;
	}

	@Override
	public void assignMap(GridMap map) {
		this.map = map;
	}

	
	@Override
	public void deleteFacts(Set<ELPLiteral> toDel) {
		if (this.bels == null)
			return;
		Iterator<InformationObject> ioIter = this.bels.iterator();
		while (ioIter.hasNext()) {
			InformationObject io = ioIter.next();
			if ( toDel.contains( io.getELP().getRule(0).getHead(0) )) {
				ioIter.remove();
			}
		}	
	}

	@Override
	public List<ELPLiteral> getExploredCells() {
		return this.newExplorations;
	}

	@Override
	public void processBeliefSet(BeliefSet bs, BeliefBase bb) {
		// update revised isopen state here!
		List<ELPLiteral> movers = new LinkedList<ELPLiteral>();
		IdentTable it = bs.getIdentifiers();
		
		for (ELPLiteral l : bs.getBeliefs()) {
			ELPAtom atm = l.getAtom();
			if (atm.getPredicate().equalsIgnoreCase("ison")) {
				boolean isAg = it.isKindOf(atm.getTerm(0), "agent");
				boolean isCow = it.isKindOf(atm.getTerm(0), "cow");
				if (isAg || isCow)
					movers.add(l);
			}
			
			if (atm.getPredicate().equalsIgnoreCase("isopen")) {
				int x = Integer.parseInt(atm.getTerm(0));
				int y = Integer.parseInt(atm.getTerm(1));
				this.map.setFlag(x, y, l);
			}
		}
		
		this.map.updateMovers(movers);
		
		// if we are in contest mode, add gates
		Set<String> processed_fences = new HashSet<String>();				
	}

	@Override
	public ELP getStateProgram() {
		return null;
	}

	@Override
	public GridMap getGridMap() {
		return this.map;
	}

	@Override
	public void reset() {
		newExplorations = new LinkedList<ELPLiteral>();
		fenceCoords = new HashSet<String>();
		switchCoords = new HashSet<String>();
	} 
			
}
