package edu.udo.cs.ie.cowbots.bdi.desgen;

import edu.udo.cs.ie.cowbots.bdi.brf.IEpistemicState;
import jason.asSyntax.Literal;

/**
 * Abstract class for any class which represents an agents 
 * desire. Each desire must have an unique name represented
 * by a Literal. When returned to other classes as needed
 * for intensity update, intensity of a desire should never
 * be less than one.
 * 
 * This interface also provides methods for intensity
 * update, one for the normal update based on the current
 * agents beliefs, and one for proof of mitigation, if 
 * the associated goal is fulfilled.
 * 
 * @author dhoelzgen
 *
 */
public abstract class CDesire implements Comparable<CDesire> {

	/**
	 * 
	 * @return The desires unique name
	 */
	public abstract Literal getUniqueName();
	
	/**
	 * 
	 * @return The desires current intensity
	 */
	public abstract int getIntensity();
	
	/**
	 * Updates the desires intensity based on the current
	 * agents beliefs.
	 * 
	 * @param Current agents beliefs for intensity update
	 * @return Updated desires intensity
	 */
	public abstract int updateIntensity(IEpistemicState bel);
	
	/**
	 * Prooves if the associated goal is fulfilled to mitigate
	 * the desires intensity. This usually sets intensity to
	 * zero to cause deletion of the desire.
	 * 
	 * @param Current agents beliefs for mitigation
	 * @return Updated desires intensity
	 */
	public abstract int updateMitigation(IEpistemicState bel);
	
	/**
	 * The equals method of all desires must be overwritten to
	 * check equality of the desires unique names.
	 * 
	 * @param Desire to check equality with 
	 * @return True, if the compared desire is equal
	 */
	@Override
	public boolean equals(Object d) {
		if (d instanceof CDesire) {
			return (this.getUniqueName().toString().equals(((CDesire)d).getUniqueName().toString()));
		}
		else return super.equals(d);
	}
}
