package edu.udo.cs.ie.cowbots.bdi.desgen;

import java.util.List;

import jason.asSyntax.Literal;
import edu.udo.cs.ie.cowbots.bdi.brf.IEpistemicState;
import edu.udo.cs.ie.cowbots.logic.ELP;

public class CowbotDesire extends DefaultDesire {

	protected String stringUpdateELP, stringMitigateELP;
	static final int cowRadius = 4;
	
	public CowbotDesire(Literal uniqueName, int i) {
		super(uniqueName, i);
	}
	
	public CowbotDesire(Literal uniqueName, int i, String updateELP, String mitigateELP) {
		super(uniqueName, i);
		this.stringUpdateELP = updateELP;
		this.stringMitigateELP = mitigateELP;
	}
	
	@Override
	public int updateIntensity(IEpistemicState bel) {
		if (this.stringUpdateELP.isEmpty())  return this.intensity;
		ELP extendedUpdateELP = new ELP();
		extendedUpdateELP.addRaw(this.additionalLiterals(bel) + this.stringUpdateELP);
		int newint = this.getNewIntensityWithELP(bel, extendedUpdateELP);
		if (newint != -1) this.intensity = newint;
		return this.intensity;
	}

	@Override
	public int updateMitigation(IEpistemicState bel) {
		if (this.stringMitigateELP.isEmpty())  return this.intensity;
		ELP extendedMitigateELP = new ELP();
		extendedMitigateELP.addRaw(this.additionalLiterals(bel) + this.stringMitigateELP);
		int newint = this.getNewIntensityWithELP(bel, extendedMitigateELP);
		if (newint != -1) this.intensity = newint;
		return this.intensity;
	}
	
	// TODO: Calculate only once for all desires
	protected String additionalLiterals(IEpistemicState bel) {
		StringBuilder addLiterals = new StringBuilder();
		
		ELP callELP = new ELP();
		callELP.addRaw("mypos(X, Y) :- ison(I, X, Y, _), iam(I).");
		List<Literal> allBeliefs = bel.getBeliefs(callELP);
		
		int agentX = 0;
		int agentY = 0;
		
		int worldTiles = 0;
		int worldW = 1;
		int worldH = 1;
		double unknownTiles = 0;
		
		for (Literal lit : allBeliefs) {
			
			if (lit.toString().startsWith("mypos(")) {
				// Get the agents position
				agentX = Integer.parseInt(lit.getTerm(0).toString());
				agentY = Integer.parseInt(lit.getTerm(1).toString());
			} else if (lit.toString().startsWith("worldsize")) {
				// Count world tiles
				worldTiles = Integer.parseInt(lit.getTerm(0).toString()) * Integer.parseInt(lit.getTerm(1).toString());
				worldW = Integer.parseInt(lit.getTerm(0).toString());
				worldH = Integer.parseInt(lit.getTerm(1).toString());
			}
		}
		
		int[][] cowWeights = new int[worldW][worldH];
		String[][] cowNames = new String[worldW][worldH];
		
		// Init arrays
		for (int i = 0; i < worldW; i++) {
			for (int j = 0; j < worldH; j++) {
				cowWeights[i][j] = 0;
				cowNames[i][j] = "NULL";
			}
		}
		
		for (Literal lit : allBeliefs) {
			
			if (lit.toString().startsWith("ison(cow")) {
				String cowName = lit.getTerm(0).toString();
				int cowX = Integer.parseInt(lit.getTerm(1).toString());
				int cowY = Integer.parseInt(lit.getTerm(2).toString());
				
				// Calculate distance from agent to cow
				String dist = String.valueOf(((int)(Math.sqrt(Math.pow(cowX - agentX, 2)+Math.pow(cowY - agentY, 2)))));
				addLiterals.append("dist(" + cowName + "," + dist + ").");
				
				try {
				// Adjust weights
				cowNames[cowX][cowY] = cowName;
				int i_floor = cowX < cowRadius ? 0 : cowX - cowRadius;
				int i_ceil = cowX > (worldW - 1 - cowRadius) ? (worldW - 1) : cowX + cowRadius;
				int j_floor = cowY < cowRadius ? 0 : cowY - cowRadius;
				int j_ceil = cowY > (worldH - 1 - cowRadius) ? (worldH - 1) : cowY + cowRadius;
				
				for (int i = i_floor; i <= i_ceil; i++) {
					for (int j = j_floor; j <= j_ceil; j++) {
						cowWeights[i][j]++;
					}
				}
				} catch (Exception ex) {
					System.err.println(ex.toString());
				
				}
				
			} else if (lit.toString().startsWith("ison(unexplored")) {
				// Count unexplored tiles
				unknownTiles++;
			}
		}
		
		// Add Cow Weight Literals
		for (int i = 0; i < worldW; i++) {
			for (int j = 0; j < worldH; j++) {
				if (cowNames[i][j] != "NULL") {
					addLiterals.append("weight(" + cowNames[i][j] + "," + String.valueOf(cowWeights[i][j]) + ").");
				}
			}
		}
		
		if (worldTiles != 0 && unknownTiles != 0)
			addLiterals.append("discovered(" + String.valueOf((int)((unknownTiles / worldTiles)*100)) + ").");
		else 
			addLiterals.append("discovered(0).");
		
		return addLiterals.toString();
	}
}
