package edu.udo.cs.ie.cowbots.bdi.desgen;

import jason.asSyntax.Literal;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;
import java.util.LinkedList;

import edu.udo.cs.ie.cowbots.bdi.brf.IEpistemicState;
import edu.udo.cs.ie.cowbots.bdi.brf.IdentTable;
import edu.udo.cs.ie.cowbots.internalActions.scout.astar;
import edu.udo.cs.ie.cowbots.logic.ELP;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;
import edu.udo.cs.ie.cowbots.util.DistanceMap;
import edu.udo.cs.ie.cowbots.util.GridMap;
import edu.udo.cs.ie.cowbots.util.PathTransformer;
import edu.udo.cs.ie.cowbots.util.cluster.ClusterCalculator;
import edu.udo.cs.ie.cowbots.util.cluster.CowCluster;
import edu.udo.cs.ie.cowbots.util.cluster.CowPosition;
import edu.udo.cs.ie.cowbots.util.cluster.IClusterCalculator;
import edu.udo.cs.ie.cowbots.util.cluster.NewClusterCalculator;
import edu.udo.cs.ie.cowbots.ui.ICowbotUI;

/**
 * Fast, java-based class for generating desires for
 * the cowbot scenario. Not really flexible, but much
 * faster than java supported dlv calls...
 * 
 * @author dhoelzgen
 *
 */
public class FastCowbotDesireComponent extends DesireComponent {
	static final int cowRadius = 4;

	static final boolean enable_scouting = true;
	static final boolean enable_cowdriving = true;
	static final boolean enable_helping = true;

	// If set to true, no drivencow Desire is generated
	// if agent has a rejectall(...) information
	static final boolean kill_all = true;
	
	// if set to false, cow cluster to corral distance is
	// maximized by regarding each cow in a cluster.
	// if set to true, the corral center is taken for
	// clustering
	protected boolean	cluster2corralUseCenter = false;
	
	// top chased cluster
	CowCluster topCluster = null;
	
	// list of all clusters
	List<CowCluster> clusters = new LinkedList<CowCluster>();
	
	ICowbotUI ui;
	
	public FastCowbotDesireComponent(ICowbotUI ui) {
		this.desireState = new DesireState();
		this.ui = ui;
	}

	/**
	 * Provides access to the agents current desirestate.
	 * 
	 * @return Current agents desirestate
	 */
	@Override
	public IDesireState getDesireState() {
		return desireState;
	}

	/**
	 * Updates the agents desirestate based on provided
	 * beliefs.
	 * 
	 * @param Current agents beliefs to update desirestate
	 * @return The updated desirestate
	 */
	@Override
	public IDesireState updateDesireState(IEpistemicState bel) {
		// Time check
		long enterTime = System.currentTimeMillis();
		long exitTime = 0;

		// Create new DesireState
		IDesireState result = new DesireState();

		// Get the required information out of the agent's beliefs
		List<Literal> allBeliefs = new ArrayList<Literal>();

		// A few hashmaps to access the information
		HashMap<String, Integer> agOrder = new HashMap<String, Integer>();
		HashMap<String, Boolean> taskKillswitch = new HashMap<String, Boolean>();
		HashMap<String, Boolean> taskAccepted = new HashMap<String, Boolean>();
		HashMap<String, Boolean> taskGotHelp = new HashMap<String, Boolean>();
		HashMap<String, Boolean> drivingCow = new HashMap<String, Boolean>();
		HashMap<String, Point> cowPositions = new HashMap<String, Point>();

		Point ownCorral = new Point(0,0);
		
		boolean foundKillswitch = false;

		String selfName = "na";

		try {
			Set<ELPLiteral> elpLits = bel.getBeliefSet().getBeliefs();

			String sLit = "";
			for (ELPLiteral lit : elpLits) {
				sLit = lit.toString();

				// Remember own name
				if (sLit.startsWith("iam(")) {
					selfName = Literal.parseLiteral(sLit).getTerm(0).toString();
					continue;
				}

				// Fill team numbers
				if (sLit.startsWith("teamnumber(")) {
					Literal orderLit = Literal.parseLiteral(sLit);
					agOrder.put(orderLit.getTerm(0).toString(), Integer.valueOf(orderLit.getTerm(1).toString()));
					continue;
				}

				// Fill accepted task list
				if (sLit.startsWith("accepted(")) {
					Literal accLit = Literal.parseLiteral(sLit);
					taskAccepted.put(accLit.getTerm(0)+","+accLit.getTerm(1)+","+accLit.getTerm(2), true);
					taskGotHelp.put(accLit.getTerm(1)+","+accLit.getTerm(2), true);
					continue;
				}

				// Remember which cow the agent is driving
				if (sLit.startsWith("drivingcow(")) {
					Literal driveLit = Literal.parseLiteral(sLit);
					drivingCow.put(driveLit.getTerm(1).toString(), true);
					continue;
				}
				
				// Set center of own corral
				if (sLit.startsWith("corralcenter(")) {
					Literal corralLit = Literal.parseLiteral(sLit);
					ownCorral = new Point(Integer.parseInt(corralLit.getTerm(0).toString()),
							Integer.parseInt(corralLit.getTerm(1).toString()));
					continue;
				}

				// Ignore some information
				if (sLit.startsWith("ison(empty")) continue;
				if (sLit.startsWith("ison(obstacle")) continue;

				// Add all other beliefs
				allBeliefs.add(Literal.parseLiteral(sLit));
			}
		} catch (Exception ex) {
			System.err.println("ERROR: Desgen could not get beliefs: "+ex);
			ex.printStackTrace();
			return this.desireState;
		}

		int agentX = 0;
		int agentY = 0;

		int ownOrderNumber = 0;
		try {
			ownOrderNumber = agOrder.get(selfName);
		} catch (Exception ex) {
			System.err.println("DESGEN ERROR: Could not get order number, please review shared knowledge.");
		}

		for (Literal lit : allBeliefs) {
			if (lit.toString().startsWith("ison("+selfName)) {
				// Get the agents position
				agentX = Integer.parseInt(lit.getTerm(1).toString());
				agentY = Integer.parseInt(lit.getTerm(2).toString());
			} else if (lit.toString().startsWith("rejectedall(")) {
				// Remember, if agent could not get help for a task
				taskKillswitch.put(lit.getTerm(2).toString(), true);
				foundKillswitch = true;
			}
		}

		for (Literal lit : allBeliefs) {

			if (lit.toString().startsWith("ison(cow")) {
				String cowName = lit.getTerm(0).toString();
				int cowX = Integer.parseInt(lit.getTerm(1).toString());
				int cowY = Integer.parseInt(lit.getTerm(2).toString());
				cowPositions.put(cowName, new Point(cowX, cowY));
			} else if (lit.toString().startsWith("help(") && enable_helping) {
				try {
					if (lit.getTerm(1).toString().equals("drivencow")) {
						// Add Helping Desire
						String agName = lit.getTerm(0).toString();
						String cowName = lit.getTerm(2).toString();

						// Check if requesting agent has a higher order number
						// TODO: Temporay disabled
						if (ownOrderNumber < agOrder.get(agName) || true) {

							int helpInt = Integer.parseInt(lit.getTerm(3).toString()); // max 85
							if (taskAccepted.containsKey(agName+",drivencow,"+cowName))
								helpInt += 80; // Hotfix, should be 15 for max 100

							result.add(new DefaultDesire(
									Literal.parseLiteral("helped(" + agName + ",drivencow," + cowName + ")"),
									helpInt));
						}
					} else {
						System.err.println("NO HELP DESIRE CREATED FOR: " + lit.toString());
					}
				} catch  (Exception ex) {
					System.err.println(ex.toString());
				}
			}
		}
		
		// check if we are a cow clusterer
		boolean doCluster = false;
		for (ELPLiteral l : bel.getBeliefSet().findAll("clustercows")) {
			if (l.getAtom().getTerm(0).equalsIgnoreCase(selfName))
				doCluster = true;
		}

		// Add Cowdriving Desires
		int currentTime = 0;
		
		// new: are we allowed to cluster this turn?
		boolean recluster = false;
		if (doCluster) {
			int leadPrioDivisor = 0;
			int leadClusterPrio = -1;
			
						
			
			ELPLiteral l = bel.getBeliefSet().findFirst("leadPrioDivisor");
			if (l != null)
				leadPrioDivisor = l.getAtom().getTermAsInt(0);
			
			l = bel.getBeliefSet().findFirst("time");
			if (l != null)
				currentTime = l.getAtom().getTermAsInt(0);
			
			for (ELPLiteral k : bel.getBeliefSet().findAll("leadClusterPrio")) {
				if (k.getAtom().getTerm(0).equalsIgnoreCase(selfName))
					leadClusterPrio = k.getAtom().getTermAsInt(1);
			}
			
			if ((leadPrioDivisor > 0) && (leadClusterPrio >= 0) && (currentTime > 0)) {
				if ((currentTime % leadPrioDivisor) == leadClusterPrio)
					recluster = true;
			}
		}
		
		if (enable_cowdriving && bel.getMap().height() > 1 && doCluster) { // TRY TO AVOID ERROR ON INIT, TEST ONLY WITH EVE
			

			
			// get cluster parameters
			int defaultClusterDistance = 4;
			
			// cluster parameters
			for (ELPLiteral l : bel.getBeliefSet().findAll("clusterdistance"))
				if (l.getAtom().getTerm(0).equalsIgnoreCase(selfName))
					defaultClusterDistance = l.getAtom().getTermAsInt(1);
			
			// corral center
			ELPLiteral corralCenter = bel.getBeliefSet().findFirst("corralcenter");
			int corralCenterX = corralCenter.getAtom().getTermAsInt(0);
			int corralCenterY = corralCenter.getAtom().getTermAsInt(1);
			
			IClusterCalculator cc = new NewClusterCalculator();
			
			// the map we operato on
			GridMap map = bel.getMap();
			// distance map
			DistanceMap dm = map.getDistanceMap(false);
			// setup fill point
			dm.fill(corralCenterX, corralCenterY);
			
			// Calculate distance from agent to cows
			DistanceMap dm2 = map.getDistanceMap(false);
			dm2.fill(agentX, agentY);
			
			//cc = new NewClusterCalculator();
			// shall we recluster?
			if (recluster) {
				
				// only cluster cows that are not older
				// than 200 turns. this will prevent agents
				// from chasing ghost cows.
				int maxAge = 200;
				int maxCowAge = (currentTime > maxAge)?currentTime-maxAge:1;
				clusters = cc.getCowClusters(map, dm, defaultClusterDistance, maxCowAge);
			}
			
			astar pathfinder = new astar();
			pathfinder.experimental_cowsight = false;
			pathfinder.experimental_obstacle = false;
			pathfinder.experimental_newposition = false;

			int topRated = 0;
			topCluster = null;
			
			for (CowCluster e : clusters) {	
				int cowX = 0;
				int cowY = 0;
				
				try {
					String cowName = "";
					int clusterDistToCorral = -1;
					int cowsInCluster = 0;
					int bestCowX = 0;
					int bestCowY = 0;
					
					if (!cluster2corralUseCenter) {
						for (CowPosition cow : e.cows) {
							cowX = cow.x;
							cowY = cow.y;
							
							int cowDistToCorral = dm.distance(cowX, cowY);
							
							if (cowDistToCorral > clusterDistToCorral) {
								clusterDistToCorral = cowDistToCorral;
								cowName = cow.name;
								bestCowX = cowX;
								bestCowY = cowY;
							}
							
							cowsInCluster++;
						}
						e.pathStartX = bestCowX;
						e.pathStartY = bestCowY;
						
						cowX = bestCowX;
						cowY = bestCowY;
						
						e.path = PathTransformer.getPath( dm.getPath(bestCowX, bestCowY));
						System.out.println("cluster path length: "+e.path.length);
						
					} else if (cluster2corralUseCenter) {
						cowX = e.centerX();
						cowY = e.centerY();
						
						String[] path2Corral = pathfinder.getPath(bel.getMap(), cowX, cowY, 
								ownCorral.x, ownCorral.y, false);
						int cowDistToCorral = path2Corral.length;
						
						if (cowDistToCorral > clusterDistToCorral) {
							clusterDistToCorral = cowDistToCorral;
							cowName = e.cows.get(0).name;
							e.path = path2Corral;
							e.pathStartX = cowX;
							e.pathStartY = cowY;
						}
					}
										
					// Do not create desires if the cow is in own corral
					boolean cowOnOwnCorral = false;
					ELPLiteral bgnd = bel.getMap().getBgnd(cowX, cowY);
					if (bgnd != null) {
						if (bgnd.toString().startsWith("ison(owncorral"))
							cowOnOwnCorral = true;
					}
					
					if (!taskKillswitch.containsKey(cowName)
							&& !(foundKillswitch && kill_all)
							&& !cowOnOwnCorral
							&& clusterDistToCorral > 0) {
												
																														
						int fetchedInt = 1; // max 1
												
						if ((bgnd != null) && (selfName.equalsIgnoreCase("pat"))) {
							if (bgnd.toString().startsWith("ison(enemycorral"))
								fetchedInt += 100;
						}
						
						
						fetchedInt += 36 - Math.min(clusterDistToCorral, 36); // max 36
						fetchedInt += (Math.min(cowsInCluster, 15) * 2) - 1; // max 29
						int distAgToCow = dm2.distance(cowX, cowY);
						fetchedInt += (distAgToCow < 20 ? (19 - distAgToCow) : 0);
						

						if (fetchedInt > topRated) {
							topRated = fetchedInt;
							topCluster = e;
						}
						result.add(new DefaultDesire(
								Literal.parseLiteral("drivencow(" + cowName + ")"),
								fetchedInt));
						
						e.acceptedCluster = true;
					}
					

				} catch (Exception ex) {
					System.err.println("Desgen Error: Could not add drivencow desire.");
					ex.printStackTrace();
				}
				
				
			}
			
			// set the top cluster as target
			if (topCluster != null)
				topCluster.targetCluster = true;

			
			// Refresh cluster view (tv)
			this.ui.updateCluster(clusters);

		}

		// Add Scouting Desires
		if (enable_scouting) {
			// TODO: Change intensity according to unknown areas on the map
			int scoutedInt = 1;
			result.add(new DefaultDesire(
					Literal.parseLiteral("scouted"),
					scoutedInt));
		}


		this.desireState = result;

		// Time check
		exitTime = System.currentTimeMillis();
		// System.out.println("Desgen Time: " + (exitTime - enterTime) + "msec");

		return desireState;
	}
	
	public List<CowCluster> getCowCLusters() {
		return Collections.singletonList(topCluster);
	}
}
