package edu.udo.cs.ie.cowbots.env;

import java.awt.*;

import javax.swing.*;

import java.util.concurrent.locks.*;

import java.util.*;

import edu.udo.cs.ie.cowbots.ui.*;

import edu.udo.cs.ie.cowbots.ui.theme.*;

/**
 * this class is used for painting massim svg output
 * onto a mappanel.
 * 
 * @author Thomas
 *
 */
public class MassimPanel extends MapPanel {

	// last svg files successfully read
	protected String				lastBgnd = "", lastTurn = "";

	// svg reader (contains cell data)
	protected MassimSVGReader		map = null, bgnd = null;
	
	// initial cell width and height (used for painting)
	protected int					cellX = 10, cellY = 10;
	
	// color table (used for painting)
	protected Color					colors[];
	
	// lock to protect prevent race conditions during paint/map update
	//protected ReentrantLock			paintlock = new ReentrantLock();
	
	/**
	 * default constructor. creates a color table
	 * 
	 * @author Thomas
	 */
	public MassimPanel() {
		super();
		colors = new Color[10];
		CowbotColorTable cct = new CowbotColorTable();
		
		colors[MassimSVGReader.Cell.T_AGENTA] = cct.getColor(CowbotColorTable.C_ALLY);
		colors[MassimSVGReader.Cell.T_AGENTB] = cct.getColor(CowbotColorTable.C_ENEMY);
		colors[MassimSVGReader.Cell.T_CORRALA] = cct.getColor(CowbotColorTable.C_CORRALOUR);
		colors[MassimSVGReader.Cell.T_CORRALB] = cct.getColor(CowbotColorTable.C_CORRALENEMY);
		colors[MassimSVGReader.Cell.T_COW] = cct.getColor(CowbotColorTable.C_COW);
		colors[MassimSVGReader.Cell.T_FENCE] = cct.getColor(CowbotColorTable.C_FENCE);
		colors[MassimSVGReader.Cell.T_SWITCH] = cct.getColor(CowbotColorTable.C_SWITCH);
		colors[MassimSVGReader.Cell.T_OBSTACLE] = cct.getColor(CowbotColorTable.C_OBSTACLE); 
	}
	

	/**
	 * prepare draw buffers before painting: update maps
	 * 
	 * @author Thomas
	 *  
	 */
	protected void prepareFrame() {
		if (bgnd != null) {
			this.setMapSize(bgnd.mapX, bgnd.mapY);
			this.autoScale = true;
			paintCells(bgnd,0);
			paintCells(map,1);
		}
	}
	
	/**
	 * paints cells from a massim svg reader data structure
	 * 
	 * @author Thomas
	 * 
	 * @param r		svg data to draw
	 * @param g		Graphics2D object
	 */
	protected void paintCells(MassimSVGReader r, int layer) {
		if (r == null)
			return;
		ArrayList<DrawCell> buffer = buffers.get(layer);
		buffer.clear();
		
		for (MassimSVGReader.Cell c : r.data) {
			buffer.add( new DrawCell(c.x, c.y, colors[c.t]));
		}
	}
	
	/**
	 * sets new svg data to read and draw
	 * 
	 * @param turnFile	step of simulation to draw
	 * @param mapFile	static simulation entities
	 */
	public void setSVG(String turnFile, String mapFile) {
		// load maps
		MassimSVGReader s = new MassimSVGReader(20,20);
		s.read(mapFile);
		
		MassimSVGReader r = new MassimSVGReader(20,20);
		r.read(turnFile);
			
		// update internal data structures
		if (s.valid && !lastBgnd.equals(mapFile)) {
			paintLock.lock();
			lastBgnd = mapFile;
			bgnd = s;
			paintLock.unlock();
		}
		
		if (r.valid && !lastTurn.equals(turnFile)) {
			paintLock.lock();
			lastTurn = turnFile;
			map = r;
			paintLock.unlock();
		}
	}
}
