package edu.udo.cs.ie.cowbots.env;

import java.io.File;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * this class reads massim 2010 svg output
 * 
 * @author Thomas
 * 
 */
public class MassimSVGReader {

	public MassimSVGReader(int x, int y) {
		this.data = new ArrayList<Cell>();
		this.scaleX = x;
		this.scaleY = y;
		this.valid = false;
	}

	/**
	 * data structure to hold cell information
	 * 
	 * @author Thomas
	 * 
	 */
	public class Cell {
		public Cell(int x, int y, int type) {
			this.x = x;
			this.y = y;
			this.t = type;
		}

		public int x, y, t;

		public static final int T_OBSTACLE = 1;
		public static final int T_SWITCH = 2;
		public static final int T_FENCE = 3;
		public static final int T_COW = 4;
		public static final int T_AGENTA = 5;
		public static final int T_AGENTB = 6;
		public static final int T_CORRALA = 7;
		public static final int T_CORRALB = 8;
	}

	// map size
	public int mapX, mapY;

	// cell data
	public ArrayList<Cell> data;

	// true if parsing svg was successful
	public boolean valid;

	// helper variables to scale down massim coordinates
	// to discrete tileworld coordinates
	protected int scaleX, scaleY;

	/**
	 * read a svg file
	 * 
	 * @author Thomas
	 * 
	 * @param file
	 */
	public void read(String file) {
		this.data.clear();
		this.valid = false;
		try {

			DocumentBuilder docbld = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			Document doc = docbld.parse(new File(file));

			NodeList l = doc.getElementsByTagName("g");
			for (int i = 0; i < l.getLength(); i++) {
				Element e = (Element) l.item(i);
				String id = e.getAttribute("id");
				if (id.equals("grid"))
					parseGrid(e);

				if (id.equals("trees"))
					parseObstacle(e);

				if (id.equals("switches"))
					parseSwitch(e);

				if (id.equals("stables"))
					parseCorral(e);

				if (id.equals("cows"))
					parseCows(e);

				if (id.equals("fences"))
					parseFences(e);

				if (id.equals("agents"))
					parseAgents(e);
			}

			this.valid = true;
		} catch (Exception e) {
			// we don't care about a specific error,
			// if map reading failed then we can
			// do nothing.
			this.valid = false;
		}
	}

	protected void parseCows(Element e) {
		addCellTransform(e, Cell.T_COW);
	}

	protected void parseFences(Element e) {
		addCellTransform(e, Cell.T_FENCE);
	}

	protected void parseAgents(Element e) {
		NodeList l = e.getElementsByTagName("use");
		for (int i = 0; i < l.getLength(); i++) {
			e = (Element) l.item(i);
			String ag = e.getAttribute("xlink:href");
			char AorB = ag.charAt(ag.length() - 1);
			// #agentred or #agentblue ?
			int t = (AorB == 'd') ? Cell.T_AGENTA : Cell.T_AGENTB;
			String tx = e.getAttribute("transform");
			data.add(this.cellTransform(tx, t));
		}
	}

	protected void parseGrid(Element e) {
		NodeList l = e.getElementsByTagName("polygon");
		if (l.getLength() > 0) {
			e = (Element) l.item(0);
			String pnt = e.getAttribute("points");
			String pnt2[] = pnt.split(" ");
			String pnt3[] = pnt2[2].split(",");
			mapX = (int) Double.parseDouble(pnt3[0]);
			mapY = (int) Double.parseDouble(pnt3[1]);
			mapX /= scaleX;
			mapY /= scaleY;
		}
	}

	protected void parseSwitch(Element e) {
		addCellTransform(e, Cell.T_SWITCH);
	}

	protected void parseObstacle(Element e) {
		addCellXY(e, Cell.T_OBSTACLE);
	}

	protected void parseCorral(Element e) {
		NodeList l = e.getElementsByTagName("use");
		for (int i = 0; i < l.getLength(); i++) {
			e = (Element) l.item(i);
			String cor = e.getAttribute("xlink:href");
			char AorB = cor.charAt(cor.length() - 1);
			int t = (AorB == '0') ? Cell.T_CORRALA : Cell.T_CORRALB;
			String sx = e.getAttribute("x");
			String sy = e.getAttribute("y");
			int x = (int) Double.parseDouble(sx);
			int y = (int) Double.parseDouble(sy);
			addCell(x, y, t);
		}
	}

	protected void addCellXY(Element e, int CellType) {
		NodeList l = e.getElementsByTagName("use");
		for (int i = 0; i < l.getLength(); i++) {
			e = (Element) l.item(i);
			String sx = e.getAttribute("x");
			String sy = e.getAttribute("y");
			int x = (int) Double.parseDouble(sx);
			int y = (int) Double.parseDouble(sy);
			addCell(x, y, CellType);
		}
	}

	protected void addCellTransform(Element e, int CellType) {
		NodeList l = e.getElementsByTagName("use");
		for (int i = 0; i < l.getLength(); i++) {
			e = (Element) l.item(i);
			String tx[] = e.getAttribute("transform").split(" ");

			// translate(X.x,Y.y)
			tx[0] = tx[0].substring(10, tx[0].length() - 1);
			tx = tx[0].split(",");
			int x = (int) Double.parseDouble(tx[0]);
			int y = (int) Double.parseDouble(tx[1]);
			addCell(x, y, CellType);
		}
	}

	protected Cell cellTransform(String s, int t) {
		String tx[] = s.split(" ");

		// translate(X.x,Y.y)
		tx[0] = tx[0].substring(10, tx[0].length() - 1);
		tx = tx[0].split(",");
		int x = (int) Double.parseDouble(tx[0]);
		int y = (int) Double.parseDouble(tx[1]);

		return new Cell(x / scaleX, y / scaleY, t);
	}

	protected void addCell(int x, int y, int t) {
		data.add(new Cell(x / scaleX, y / scaleY, t));
	}

	public String toString() {
		return "map: " + mapX + "x" + mapY + ", " + data.size() + " elements";
	}
}
