package edu.udo.cs.ie.cowbots.internalActions.experimental2;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;

import edu.udo.cs.ie.cowbots.agent.ICowbot;
import edu.udo.cs.ie.cowbots.bdi.brf.BeliefSet;
import edu.udo.cs.ie.cowbots.internalActions.scout.astar;
import edu.udo.cs.ie.cowbots.logic.ELPAtom;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;
import edu.udo.cs.ie.cowbots.ui.ICowbotUI;
import edu.udo.cs.ie.cowbots.util.DistanceMap;
import edu.udo.cs.ie.cowbots.util.GridMap;

import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.ASSyntax;
import jason.asSyntax.Literal;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.NumberTermImpl;
import jason.asSyntax.Plan;
import jason.asSyntax.Pred;
import jason.asSyntax.Term;

/**
 * Faster replacement for the dynamic planner. Not really adjustable...
 * 
 * @author dhoelzgen
 *
 */
public class exCreatePlan extends DefaultInternalAction {

	HashMap<Point, Point> gatePositions = new HashMap<Point, Point>();
	HashMap<Point, String> gateDirections = new HashMap<Point, String>();
	
	// map used to represent leader <-> helper lists
	// the key is a leader name, the value is a list of helpers (leader always position 0)
	HashMap<String, List<String>> groups = new HashMap<String, List<String>>();
	
	// map used to represent agent positions
	// the key is an agents name, the value a point
	HashMap<String, Point> agPositions = new HashMap<String, Point>();
	List<String> agHelpers = new LinkedList<String>();
	HashMap<String,Integer> agIndex = new HashMap<String,Integer>();
	HashMap<Integer,String> agInvIndex = new HashMap<Integer,String>();
	
	// the agents name in the mas
	String myself;
	
	// IgnoreList
	ArrayList<Integer> ignoreList = new ArrayList<Integer>();
	boolean startWithIgnore = false;
	
	// used for logging
	List<String>	debugLog;
		
	/**
	 * Used by a teamleader to generate an asl plan to send orders to the team.
	 * Generates a task to open the next fence, and send the others driving the cow.
	 * Can ignore agents, if there is no way to get them to the group, or if
	 * they have to wait because of a switch which is only accessible from one
	 * side of the fence.
	 */
	@Override
	public Object execute(TransitionSystem ts, Unifier un, Term[] args) throws Exception {
		try {
			// Cows position
			int cX = (int) ((NumberTerm) args[0]).solve();
			int cY = (int) ((NumberTerm) args[1]).solve();
			
			Point cowPos = new Point(cX, cY);

			// Targets position
			int tX = (int) ((NumberTerm) args[2]).solve();
			int tY = (int) ((NumberTerm) args[3]).solve();
			
			Point targetPos = new Point(tX, tY);

			// The agents positions
			
			Point[] aPositions;
			
			// Fill gate positions
			this.fillGatePositions(ts);

			// Init plan
			String rPlan = "+!sendOrders : true <-"
				+ "?kwaypoint(startwp,S,X,Y,F,D);"
				+ "?drivingCow(C);"
				+ "?time(Q);";
			
			// IgnoreList Handling
			if (!startWithIgnore) {
				this.ignoreList = new ArrayList<Integer>();
			}
			this.startWithIgnore = false;

			// cbdi extensions
			ICowbot cowbot = (ICowbot) ts.getAg();
			BeliefSet bs = cowbot.getEpistemicState().getBeliefSet();
			GridMap map = cowbot.getEpistemicState().getMap();
			ICowbotUI ui = cowbot.getUI();
			this.debugLog = new LinkedList<String>();
			
			// agents name
			myself = ts.getUserAgArch().getAgName();
			
			// build groups and ag Positions
			this.logSeparator("lead: "+myself);
			this.buildGroups(bs);
			
			
			//tv: this makes the helper list dynamic
			if (agInvIndex.size() == 0) {
				System.err.println(this.myself +" exPlanner warning: helper list is empty!");
			}
			aPositions = new Point[agInvIndex.size()];
			
			for (int i = 0; i < agInvIndex.size(); i++) {
				String agKey = agInvIndex.get(i);
				Point agPos =agPositions.get(agKey);
				aPositions[i] = agPos;
			}			
			
			// Init pathfinder (with agents and obstacles)
			DistanceMap pathfinder = map.getDistanceMap(false, false);

			// Phase One: Check for fences between agents and cows
			pathfinder.fill(cX, cY);
			
			// List with fences for each agent
			int[] fenceCountOne = new int[aPositions.length];
			int maxFenceCount = 0;
			
			// List with agents, with max number of fences in their way
			ArrayList<Integer> maxFenceCountAgents = new ArrayList<Integer>(); 

			Point nextFencePosition = new Point();
			
			this.logSeparator("fences on way");
			
			// Get next fence between agents and cow
			// Remember the number of fences in the way of each agent
			for (int i = 0; i < aPositions.length; i++) {
				if (ignoreList.contains(i)) 
				{
					String aname = this.agInvIndex.get(i);
					this.debugLog.add(aname + ": ignored");
					continue;
				}
				Point aPos = aPositions[i];
				LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
				int currentFenceCount =
					this.getNumberOfFencesOnPath(agPath, map);

				fenceCountOne[i] = currentFenceCount;

				String aname = this.agInvIndex.get(i);
				this.debugLog.add(aname + ": "+ currentFenceCount);
				
				if (currentFenceCount > maxFenceCount) {
					maxFenceCount = currentFenceCount;
					nextFencePosition = this.getNextFencePositionOnPath(agPath, map);
					maxFenceCountAgents.clear();
					maxFenceCountAgents.add(i);
				} else if (currentFenceCount == maxFenceCount) {
					maxFenceCountAgents.add(i);
				}
			}
			
			this.debugLog.add("Agents with max count: "+ maxFenceCountAgents.size());
			
			int nearestAgent = -1;
			int nearestDist = Integer.MAX_VALUE;

			if (maxFenceCount > 0) {
				
				if (maxFenceCountAgents.size() > 1) {
					// Open door with nearest agent
					for (int i = 0; i < aPositions.length; i++) {
						Point aPos = aPositions[i];
						Point gPos = this.getSwitchPosition(bs, aPos, nextFencePosition, map);
						if (gPos == null) continue;
						pathfinder = map.getDistanceMap(false, true);
						pathfinder.fill(gPos.x, gPos.y); // DH NULL POINTER EXP HERE (ContestMapTwo)
						LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
						if (agPath != null) {
							int newDist = agPath.size();
							if (newDist < nearestDist) {
								nearestAgent = i;
								nearestDist = newDist;
							}
						} else {
							this.debugLog.add("case 1: no path to reach switch position");
						}
					}
					
					this.debugLog.add("case 1: nearest agent: " + this.agInvIndex.get(nearestAgent));
					
				} else if (maxFenceCountAgents.size() == 1) {
					// Open door with nearest agent, who is not the one with
					// the most gates in its way
					for (int i = 0; i < aPositions.length; i++) {
						if (!maxFenceCountAgents.contains(i)) {
							Point aPos = aPositions[i];
							Point gPos = this.getSwitchPosition(bs, aPos, nextFencePosition, map); 
							if (gPos == null) continue;
							pathfinder = map.getDistanceMap(false, true);
							pathfinder.fill(gPos.x, gPos.y);
							LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
							if (agPath != null) {
								int newDist = agPath.size();
								if (newDist < nearestDist) {
									nearestAgent = i;
									nearestDist = newDist;
								}
							}
						}
					}
					
					if (nearestAgent == -1) {
						// Switch is only accessible from one side
						// Ignore this agent in this plan
						this.startWithIgnore = true;
						this.ignoreList.add(maxFenceCountAgents.get(0));
						this.debugLog.add("case 2: ignoring agent: " + this.agInvIndex.get(maxFenceCountAgents.get(0)));
						return this.execute(ts, un, args); // Gool luck next time ;)
					} else this.debugLog.add("case 2: nearest agent: " + this.agInvIndex.get(nearestAgent));
				}
			}
			
			// Special case handling
			pathfinder = map.getDistanceMap(false, false);
			pathfinder.fill(tX, tY);
			LinkedList<Point> tmpCowPath = pathfinder.getPath(cX, cY);
			Point tmpNextFencePosition = this.getNextFencePositionOnPath(tmpCowPath, map);
			
			// Other case for phase two
			// New if construct, to handle special case with last fence and
			// own corral
			if (maxFenceCount == 0
					|| (maxFenceCount == 1 && this.isSameSwitch(nextFencePosition, tmpNextFencePosition))) {
				// Phase Two: Check for fences between cows and target

				pathfinder = map.getDistanceMap(false, false);
				pathfinder.fill(tX, tY);
				LinkedList<Point> cowPath = pathfinder.getPath(cX, cY);
				nextFencePosition = this.getNextFencePositionOnPath(cowPath, map);

				if (nextFencePosition != null) {
					// Found a gate, find the nearest agent
					for (int i = 0; i < aPositions.length; i++) {
						Point aPos = aPositions[i];
						Point gPos = this.getSwitchPosition(bs, aPos, nextFencePosition, map);
						if (gPos == null) continue;
						pathfinder = map.getDistanceMap(false, true);
						pathfinder.fill(gPos.x, gPos.y);
						LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
						if (agPath != null) {
							int newDist = agPath.size();
							if (newDist < nearestDist) {
								nearestAgent = i;
								nearestDist = newDist;
							}
						}
					}
					
					this.debugLog.add("case 3: nearest agent: " + this.agInvIndex.get(nearestAgent));
				}

			}
			
			// Create the agents tasks
			int formPosCounter = 1;
			for (int i = 0; i < aPositions.length; i++) {
				// tv: planing agent should use + and not tell for belief additions.
				// otherwise, he want to trigger a new plan without having received
				// the believe via tell..
				//if (i == aPositions.length - 1) {
				if (this.groups.containsKey(this.getAgentName(i))) {
					
					// Leader
					if (i == nearestAgent) {
						rPlan += "+taskopenfence("+ nextFencePosition.x + "," + nextFencePosition.y + ",Q);";
					} else if (ignoreList.contains(i)) {
						// Agent is ignored, send him to his fence
						Point aPos = aPositions[i];
						pathfinder = map.getDistanceMap(false, false);
						pathfinder.fill(cX, cY);
						LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
						Point igFencePos = this.getNextFencePositionOnPath(agPath, map);
						rPlan += "+taskopenfence("+ igFencePos.x + "," + igFencePos.y + ",Q);";
					} else {
						rPlan += "+waypoint(startwp,drivencow,C,S,X,Y,F,D," + formPosCounter + ",5,Q);";
					}
					
				} else {
					
					// Helper agents
					if (i == nearestAgent) {
						rPlan += ".send("+ this.getAgentName(i) + ",tell,taskopenfence("
						+ nextFencePosition.x + "," + nextFencePosition.y + ",Q));";
					} else if (ignoreList.contains(i)) {
						// Agent is ignored, send him to his fence
						Point aPos = aPositions[i];
						pathfinder = map.getDistanceMap(false, false);
						pathfinder.fill(cX, cY);
						LinkedList<Point> agPath = pathfinder.getPath(aPos.x, aPos.y);
						Point igFencePos = this.getNextFencePositionOnPath(agPath, map);
						rPlan += ".send("+ this.getAgentName(i) + ",tell,taskopenfence("
						+ igFencePos.x + "," + igFencePos.y + ",Q));";
					} else {
						rPlan += ".send(" + this.getAgentName(i)
						+ ",tell,waypoint(startwp,drivencow,C,S,X,Y,F,D," + formPosCounter + ",5,Q));";
						formPosCounter++;						
					}
				}
			}

			// Add end to generated plan
			rPlan += ".print(\"End of generated plan for sending orders.\").";

			// Remove old plan
			while (ts.getAg().getPL().get("sendOrders") != null) {
				ts.getAg().getPL().remove("sendOrders");	
			}
			
			System.err.println("NEW PLAN " + rPlan);

			// Add new plan
			Plan newPlan = ASSyntax.parsePlan(rPlan);
			Pred p = new Pred("sendOrders");
			newPlan.setLabel(p);
			ts.getAg().getPL().add(newPlan);

			ui.updateExPlanner(this.debugLog);
			
		} catch (Exception ex) {
			System.err.println("Ex2 Create Plan: An error occured");
			ex.printStackTrace();
		}

		return true;
	}

	/**
	 * Returns the number of fences known on the provided path.
	 * 
	 * @param path
	 * @param map
	 * @return
	 */
	protected int getNumberOfFencesOnPath(LinkedList<Point> path, GridMap map) {
		int result = 0;
		boolean lastWasFence = false;

		if (path != null) {
			for (Point step : path) {
				// Search for a fence
				if (this.isFenceOnPosition(step.x, step.y, map)) {
					if (!lastWasFence) result++;
					lastWasFence = true;
				} else lastWasFence = false;
			}
		}

		return result;
	}

	/**
	 * Returns the next fence position on the provided path, or at least
	 * a point where the agent should look for a fence
	 * 
	 * @param path
	 * @param map
	 * @return
	 */
	protected Point getNextFencePositionOnPath(LinkedList<Point> path, GridMap map) {
		if (path == null) return null;
		
		for (Point step : path) {
			// Search for a fence
			if (this.isFenceOnPosition(step.x, step.y, map)) {
				return step;
			}
		}

		return null;
	}

	/**
	 * Returns the agents name, used to send orders in jason
	 * 
	 * @param i
	 * @return
	 */
	protected String getAgentName(int i) {
		return agInvIndex.get(i);
	}

	/**
	 * Fills the gate positions according to the agents beliefs
	 * 
	 * @param ts
	 */
	protected void fillGatePositions(TransitionSystem ts) {
		ICowbot cowbot = (ICowbot) ts.getAg();

		for (ELPLiteral lit : cowbot.getEpistemicState().getBeliefSet().getBeliefsByPredicate("gate")) {
			
			String[] gateString = lit.getAtom().getTerm(2).toString().split("_");

			int fX = lit.getAtom().getTermAsInt(0);
			int fY = lit.getAtom().getTermAsInt(1);

			int gX = Integer.parseInt(gateString[1]);
			int gY = Integer.parseInt(gateString[2]);

			this.gatePositions.put(new Point (fX, fY), new Point(gX, gY));
			this.gateDirections.put(new Point (fX, fY), lit.getAtom().getTerm(3).toString());
		}
	}
	
	/**
	 * Returns true, if both provided fence tiles belong to the same fence
	 * 
	 * @param f1
	 * @param f2
	 * @return
	 */
	protected boolean isSameSwitch(Point f1, Point f2) {
		Point g1 = gatePositions.get(f1);
		Point g2 = gatePositions.get(f2);
		
		if (g1 != null && g2 != null) {
			if (g1.equals(g2)) return true;
		}
		
		return false;
	}
	
	/**
	 * Returns the position of the switch. If the switch is unknown, return a
	 * good point for looking for the switch
	 * 
	 * @param aPos Agents position
	 * @param sPos Position of the switch
	 * @param direction Direction of the switch
	 * @return
	 */
	public Point getSwitchPosition(BeliefSet bs, Point aPos, Point fencePosition, GridMap map) {

		Point sPos = gatePositions.get(fencePosition);
		
		if (sPos != null) {
			
			// Known switch
			String direction = gateDirections.get(fencePosition);
			return this.moveSwitchPointToAgent(aPos, sPos, direction, map);
			
		} else {
			
			// Unknown switch
			List<ELPLiteral> fenceGroups = bs.findAll("fencegroup");
			String direction = "";
			
			for (ELPLiteral groupEntry : fenceGroups) {
				int gX = groupEntry.getAtom().getTermAsInt(0);
				int gY = groupEntry.getAtom().getTermAsInt(1);
				
				if (gX == fencePosition.x || gY == fencePosition.y) {
					// Literal is a part of the selected fence
					if (direction.isEmpty()) direction = groupEntry.getAtom().getTerm(2).toString();
					
					// Check for unknown fields
					try {
						// BUG: Fencegroup does not contain the fence at the edge
						// If changed, set 2 to 1
						if (direction.equals("horizontal")) {
							if (map.getContent(gX + 2, gY).getAtom().getTerm(0).contains("unexplored"))
								return this.moveSwitchPointToAgent(aPos, new Point(gX + 2, gY), direction, map);
							if (map.getContent(gX - 2, gY).getAtom().getTerm(0).contains("unexplored"))
								return this.moveSwitchPointToAgent(aPos, new Point(gX - 2, gY), direction, map);
						} else {
							if (map.getContent(gX, gY + 2).getAtom().getTerm(0).contains("unexplored"))
								return this.moveSwitchPointToAgent(aPos, new Point(gX, gY + 2), direction, map);
							if (map.getContent(gX, gY - 2).getAtom().getTerm(0).contains("unexplored"))
								return this.moveSwitchPointToAgent(aPos, new Point(gX, gY - 2), direction, map);
						}
					} catch (Exception ex) {
						// TODO: Decide how to handle exception
						System.err.println("Error in Switchfinder.");
						ex.printStackTrace();
					}
				}
			}
			return null;
		}
	}
	
	/**
	 * Used to get the switch position on the agents side on the fence
	 * 
	 * @param aPos
	 * @param tPos
	 * @param direction
	 * @param map
	 * @return
	 */
	protected Point moveSwitchPointToAgent(Point aPos, Point tPos, String direction, GridMap map) {
		Point result = new Point(tPos.x, tPos.y);
		if (direction.equals("horizontal")) {
			if (tPos.y < aPos.y) result.y++;
			else result.y--;
		} else {
			if (tPos.x < aPos.x) result.x++;
			else result.x--;
		}
		
		// Check, if created point is an obstacle
		if (map.getCell(result.x, result.y).isObstacle
				|| map.getCell(result.x, result.y).isSwitch
				|| map.getCell(result.x, result.y).isFence)
			return null;
		
		return result;
	}
	
	protected boolean isFenceOnPosition(int x, int y, GridMap map) {
		return map.getCell(x, y).isFence;
	}
	
	/**
	 * this method sets up some internal data structure for planning,
	 * in this case agent positions and helper lists. the groups list
	 * is based on the beliefs of an agent.
	 * @param bs
	 */
	protected void buildGroups(BeliefSet bs) {
		List<ELPLiteral> leaders = bs.findAll("leader");
		List<ELPLiteral> helpers = bs.findAll("helper");
		List<ELPLiteral> positions = bs.findAll("agentposition");
		Set<String> validHelper = new HashSet<String>();
		
		for (ELPLiteral l : positions) {
			String vh = l.getAtom().getTerm(0);
			if (vh.equalsIgnoreCase("self"))
				vh = myself;
			validHelper.add(vh);
		}
		
		String ldText = "";
				
		agHelpers.clear();
		
		// all leaders
		for (ELPLiteral i : leaders) {
			List<String> j = new LinkedList<String>();
			String ld = i.getAtom().getTerm(0);
			this.groups.put(i.getAtom().getTerm(0), j);
			if (ld.equalsIgnoreCase(this.myself))
				j.add(this.myself);
			ldText += ld+"  "; 
		}
		
		// all helpers with their leader group
		for (ELPLiteral i : helpers) {
			String leadName = i.getAtom().getTerm(0);
			String helpName = i.getAtom().getTerm(1);
			
			List<String> j = this.groups.get(leadName);
			if (j != null)
				j.add(helpName);
		}
		
		
		// prepare output		
		logSeparator("leaders");
		debugLog.add( ldText );		
		logSeparator("helpers");
		for (Entry<String, List<String>> es : groups.entrySet()) {
			String t = es.getKey() + ": ";
			for (String val : es.getValue())
				t += val + " ";
			debugLog.add(t);
		}
		
		// get my helper list
		List<String> myHelpers = groups.get(myself);
		
		// sort that array
		Collections.sort(myHelpers);
		
		// get agent positions		
		logSeparator("my helpers");
		for (ELPLiteral l : positions) {
			ELPAtom atm = l.getAtom();
			String name = atm.getTerm(0);
			if (name.equalsIgnoreCase("self"))
				name = this.myself;
			Point pos = new Point(atm.getTermAsInt(1), atm.getTermAsInt(2)); 
			this.agPositions.put( name, pos ); 			
		}
		
		this.agIndex.clear();
		this.agInvIndex.clear();
		int i = 0;
		for (String ag : myHelpers ) {
			if (validHelper.contains(ag)) {
				this.agIndex.put( ag, i);
				this.agInvIndex.put( i, ag);
				++i;
			}
		}
	
		// now the aPositions can be setup
		for (int j = 0; j < agInvIndex.size(); j++) {
			String agName=agInvIndex.get(j);			
			Point pos = this.agPositions.get( agName );
			if ((agName != null ) && (pos != null))
				debugLog.add(j+": "+agName + " at ("+pos.x+","+pos.y+")");
		}
	}
	
	
	/**
	 * this method adds a separator (line of "-" to an agents debug log
	 */
	protected void logSeparator(String caption) {
		int sepLen = 20;
		
		if (caption != null) {
			sepLen -= caption.length() -2;
			if (sepLen < 0)
				sepLen = 0;
			String logEntry = "\n-- " + caption + " ";
			for (int i = 0; i < sepLen; i++)
				logEntry += "-";
			debugLog.add(logEntry);
		} else {
			String logEntry = "\n----------------------------";
			debugLog.add(logEntry);
		}
	}
}
