package edu.udo.cs.ie.cowbots.internalActions.experimental2;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;

import edu.udo.cs.ie.cowbots.bdi.brf.BeliefBase;
import edu.udo.cs.ie.cowbots.bdi.brf.BeliefSet;
import edu.udo.cs.ie.cowbots.bdi.brf.IKnowledgeOperator;
import edu.udo.cs.ie.cowbots.bdi.brf.InformationObject;
import edu.udo.cs.ie.cowbots.logic.ELP;
import edu.udo.cs.ie.cowbots.logic.ELPAtom;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;

/**
 * 
 * knowledge operator for the mapc contest, as a fall back
 * if we are running into dlv execution time problems
 * 
 * @author Thomas
 *
 */
public class exKnowledge implements IKnowledgeOperator {
	
	HashMap<String, ELPLiteral> agentposition;
	HashMap<String, Map<String, ELPLiteral>> miniBB;
	
	public exKnowledge() {
		init();
	}
	
	protected void init() {		
		agentposition = new HashMap<String, ELPLiteral>();		
		miniBB = new HashMap<String, Map<String, ELPLiteral>>();
		miniBB.put("agentposition_msg", agentposition);
	}
	
	@Override
	public void deleteFacts(Set<ELPLiteral> toDel) {
		// TODO Auto-generated method stub

	}

	@Override
	public ELP getStateProgram() {
		// do not add rules here, dlv might not be
		// used at all
		return null;
	}

	@Override
	public void init(Element config) {
		// TODO Auto-generated method stub

	}

	@Override
	public void processBeliefSet(BeliefSet bs, BeliefBase base) {
		ELPAtom atm;
		
		// process all helperposition_msg rules
		for (ELPLiteral l : agentposition.values()) {
			//bs.addBelief( new ELPAtom("helperposition", l.getAtom().getTerms()));
			bs.addBelief( new ELPAtom("agentposition",l.getAtom().getTerm(0),
					l.getAtom().getTerm(1), l.getAtom().getTerm(2) ));
		}
		
		// process agents own position
		if (base.agIson != null) {
			atm = base.agIson.getAtom();
			bs.addBelief( new ELPAtom("agentposition",
					atm.getTerm(0),
					atm.getTerm(1),
					atm.getTerm(2) ));			
		}
		
		// gate are processed by map update in contest mode		
		// fencegroups are processed by map update in contest mode
		//
		// add agentposition as ison, so they can be recognized
		// as obstacles by the pathfinder.
		List<ELPLiteral> mymovers = new LinkedList<ELPLiteral>();
		for (ELPLiteral l : agentposition.values()) {
			mymovers.add(new ELPAtom("ison","ally",l.getAtom().getTerm(1),
					l.getAtom().getTerm(2),"1"));
		}		
		base.getMapOperator().getGridMap().updateMovers(mymovers);
	}
	

	@Override
	public boolean update(List<InformationObject> updates, BeliefBase bb) {
		
		HashMap<String, ELPLiteral> bbentry;
		
		for (InformationObject io : updates) {
			ELPLiteral fact = io.getELP().getRule(0).getHead(0);
			ELPAtom atm = fact.getAtom();
			String functor = atm.getPredicate();
			
			Map<String, ELPLiteral> bbEntry;
			
			// add helperposition_msg and helperposition here
			if (functor.equals("agentposition_msg")) {
				
				/*
				:- agentposition_msg(I,A,B,X), agentposition_msg(I,C,D,Y), X<Y.
				agentposition(I,A,B) :- agentposition_msg(I,A,B,_).
				*/
				String ag = atm.getTerm(0);
				bbEntry = miniBB.get(functor);
				ELPLiteral previous = bbEntry.get(ag);
				fact = newerFact(3,fact,previous);
				bbEntry.put(ag,fact);
			} else {
				// predicate symbol not handled, just route
				// the data into void.
			}
			
		}

		return false;
	}

	/**
	 * this method compares two literals based on their time-value
	 * encoded in a term argument.
	 * 
	 * @param timeTerm	index of time term
	 * @param a		literal a
	 * @param b		literal a or null
	 * @return		a or b (whichever has higher time term)
	 */
	protected ELPLiteral newerFact(int timeTerm, ELPLiteral a, ELPLiteral b) {
		if (b == null)
			return a;
		
		int timeA = a.getAtom().getTermAsInt(timeTerm);
		int timeB = a.getAtom().getTermAsInt(timeTerm);
		if (timeA >= timeB)
			return a;
		else
			return b;
	}

	@Override
	public void reset() {
		init();
	}
}
