package edu.udo.cs.ie.cowbots.internalActions.experimental2;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;

import edu.udo.cs.ie.cowbots.agent.ICowbot;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;
import edu.udo.cs.ie.cowbots.util.DistanceMap;
import edu.udo.cs.ie.cowbots.util.GridMap;
import edu.udo.cs.ie.cowbots.util.PathTransformer;
import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.Atom;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.Term;
import jason.stdlib.print;

/**
 * 
 * @author dhoelzgen
 *
 */
public class exNewWayfinder extends DefaultInternalAction {
	
	static String[] dirs = {"west","northwest","southwest","south","north","east","northeast","southeast"};
	        
	/**
	 * New wayfinder, using the new DistanceMap instead of a*, which either had to many
	 * bugs or was simply inappropriate. Returns the next step to the provided target.
	 * If the target is reached a skip command is returned, if it's impossible
	 * to reach, it returns a random action (turned out that this is better than
	 * do nothing)
	 */
	@Override
	public Object execute(TransitionSystem ts, Unifier un, Term[] args)
	throws Exception {
		ICowbot cowbot = (ICowbot) ts.getAg();
		GridMap map = cowbot.getEpistemicState().getMap();
		DistanceMap pathfinder = map.getDistanceMap(true);
		
		int sX = (int) ((NumberTerm) args[0]).solve();
		int sY = (int) ((NumberTerm) args[1]).solve();
		int tX = (int) ((NumberTerm) args[2]).solve();
		int tY = (int) ((NumberTerm) args[3]).solve();
		
		// Use target as fill start
		pathfinder.fill(tX, tY);
		
		// Use starting point from agent to move
		LinkedList<Point> path = pathfinder.getPath(sX, sY);
		
		String action = "skip";
		
		if (path != null) {
			// first point on path is always the starting point
			// so we always have to use 2nd entry to get a direction
			// for moving.
			if (path.size() > 1) {
				action = getMove(path.get(0), path.get(1));
				cowbot.getUI().updateAgentPath(path);
			}
		} else {
			int r = (int) (Math.random() * 8);
			action = dirs[r];
		}
		
		
		
		System.err.println(ts.getUserAgArch().getAgName()+" path action: "+action);
		
		
		return un.unifies(args[4], new Atom(action));
	}

	private String getMove(Point from, Point to) {
		return PathTransformer.getMove(from, to);
	}
}
