package edu.udo.cs.ie.cowbots.internalActions.scout;

import java.util.List;

import edu.udo.cs.ie.cowbots.agent.ICowbot;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;
import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.NumberTermImpl;
import jason.asSyntax.Term;

/**
 * Class for calculating the distance of relevant Helpers to a Closewaypoint.
 * Called by Leaders exclusively to determine whether all his Groupmembers
 * have passed a fence or not.
 * 
 * @author bienek, jablkowski
 *
 */
public class checkHelpersNearCWP extends DefaultInternalAction {

	ICowbot cowbot;

	public Object execute(TransitionSystem ts, Unifier un, Term[] args)
			throws Exception {

		cowbot = (ICowbot) ts.getAg();
		int xcoord = (int) ((NumberTerm) args[0]).solve();
		int ycoord = (int) ((NumberTerm) args[1]).solve();
		String fenceclosername = args[2].toString();

		int result = checkHelpers(xcoord, ycoord, fenceclosername);

		un.unifies(args[3], new NumberTermImpl(result));
		return true;
	}

	/**
	 * Utility function for calculating whether all relevant Helpers are near 
	 * the Closewaypoint
	 * 
	 * @param x	X-coordinate of the Waypoint
	 * @param y	Y-coordinate of the Waypoint
	 * @param name	the name of the helper responsible of keeping the door open, 
	 * the leader has to make sure he doesnt include that helper in his calculation
	 * @return	result-int, value = 1, if all relevant helpers are nearby, value = 0 else
	 *
	 */
	private int checkHelpers(int x, int y, String name) {
		List<ELPLiteral> helperpositions = cowbot.getEpistemicState()
				.getBeliefSet().getBeliefsByPredicate("helperison");
		//System.out.println("***********************************************   "+helperpositions);
		for (ELPLiteral e : helperpositions) {
			if (e.getAtom().getTerm(0). equals(name) == false) {
				int Agxcoord = Integer.parseInt(e.getAtom().getTerm(1));
				int Agycoord = Integer.parseInt(e.getAtom().getTerm(2));
				int value = calculateDistance(Agxcoord, Agycoord, x, y);
				System.out.println(e.getAtom().getTerm(0) + " "+ value);
				if (value > 2){
					
					return 0;
					}
			}
		}
		//System.out.println("************************************************************* 111111111111111111 ");
		return 1;
	}

	private int calculateDistance(int myX, int myY, int destX, int destY) {
		int diffX = Math.abs(myX - destX);
		int diffY = Math.abs(myY - destY);

		return diffX + diffY;
	}
}
