package edu.udo.cs.ie.cowbots.internalActions.scout;

import java.awt.Point;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import edu.udo.cs.ie.cowbots.agent.ICowbot;
import edu.udo.cs.ie.cowbots.logic.ELPLiteral;
import edu.udo.cs.ie.cowbots.planer.makePlan;
import edu.udo.cs.ie.cowbots.util.GridMap;
import jason.*;
import jason.asSyntax.*;
import jason.asSemantics.*;

import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.IntendedMeans;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;

/**
 * Class to calculate the Helper, who should be ordered to serach for a switch
 * if the switch for the relevant fence isn't found yet
 * 
 * @author bienek, jablkowski
 *
 */

public class getScoutSwitchFinder extends DefaultInternalAction {
     
	
	ICowbot cowbot;
	Point target1;
	Point target2;
	GridMap map;
		
	@Override
	public Object execute(TransitionSystem ts, Unifier un, Term[] args)
			throws Exception {
		   
		int  wpxcoord = (int) ((NumberTerm) args[0]).solve();
	 	int  wpycoord = (int) ((NumberTerm) args[1]).solve();		
		int  fxcoord = (int) ((NumberTerm) args[2]).solve();
	 	int  fycoord = (int) ((NumberTerm) args[3]).solve();
	 	String dir =  args[4].toString();
	 	cowbot = (ICowbot) ts.getAg();
	 	map = cowbot.getEpistemicState().getMap();
	 		 	
	 	
	 	settargets(fxcoord, fycoord,wpxcoord, wpycoord, dir);
	 	un.unifies(args[6], new NumberTermImpl(target1.x));
		un.unifies(args[7], new NumberTermImpl(target1.y));
		un.unifies(args[9], new NumberTermImpl(target2.x));
		un.unifies(args[10], new NumberTermImpl(target2.y));
		
		String[] names = calcfinders(target1, target2);
		un.unifies(args[5], new Atom(names[0]));
		un.unifies(args[8], new Atom(names[1]));
		
		
		return true;
	}
	
	/**
	 * Utility function for setting the targets the searching helpers should go to.
	 * @param xcoord X-coordinate of the fencecell, which belongs to the fence
	 * @param ycoord Y-coordinate of the fencecell, which belongs to the fence
	 * @param wpxcoord X-coordinate waypoint required to calculate the coordinates (usually the fenceopenwp)
	 * @param wpycoord Y-coordinate waypoint required to calculate the coordinates (usually the fenceopenwp)
	 * @param dir direction of the fence (vertical or horizontal)
	 * @return	Point encoding the coordinates
	 */
	private void settargets(int xcoord, int ycoord, int wpxcoord, int wpycoord, String dir){
	    if (dir.equals("vertical")) {
	    	int index = xcoord;
	    	int indey = ycoord+1;
	    	while (map.getContent(index, indey).getAtom().getTerm(0).equals("fence")) {
	    		indey++;
	    	}
	    	target1 = new Point(wpxcoord, indey-1);
	    	
	    	indey = ycoord-1;
	    	while (map.getContent(index, indey).getAtom().getTerm(0).equals("fence")) {
	    		indey--;
	    	}
	    	target2 = new Point(wpxcoord, indey+1);
	    }
	    else{
	    	//dir horizontal
	    	int index = xcoord+1;
	    	int indey = ycoord;
	    	while (map.getContent(index, indey).getAtom().getTerm(0).equals("fence")) {
	    		index++;
	    	}
	    	target1 = new Point(index-1, wpycoord);
	    	
	    	index = xcoord-1;
	    	while (map.getContent(index, indey).getAtom().getTerm(0).equals("fence")) {
	    		index--;
	    	}
	    	target2 = new Point(index+1, wpycoord);
	    }
	    
	    
	}
	
	
	/**
	 * Utility function to get the two helpers required to search for the switch
	 * @param p1 first destination, where to look
	 * @param p2 second destination, where to look
	 * @return	array containing both helper-names
	 */
	
	private String[] calcfinders(Point p1, Point p2){
		String[] strarr = new String[2];
		strarr[0] =  getHelper (p1.x,p1.y,"");
		strarr[1] =  getHelper (p2.x,p2.y,strarr[0]);
		
		return strarr;
		}
	
	/**
	 * Utility function for calculating the distance between Agent and point to go
	 * @param e the ELPLiteral encoding the agents position.
	 * @param x	X-coordinate of the point
	 * @param y	Y-coordinate of the point
	 * @return	the distance
	 *
	 */
	private int calcDistance (ELPLiteral e, int x, int y){
		int diffX = Math.abs(x-Integer.parseInt(e.getAtom().getTerm(1)));
		int diffY = Math.abs(y-Integer.parseInt(e.getAtom().getTerm(2)));
		
		
		return diffX+diffY;
	}
	
	/**
	 * Utility function to get the nearest Helper to the point specified
	 * @param x	X-coordinate of the pointCP
	 * @param y	Y-coordinate of the point
	 * @param name the name to be excluded in the search (to avoid that the same agent is chosen twice)
	 * @return	the agent's name
	 *
	 */
	private String getHelper(int x, int y, String name){
		List<ELPLiteral> helperpositions = cowbot.getEpistemicState().getBeliefSet().getBeliefsByPredicate("helperison");
		
		if (helperpositions.size() != 0){
		int nearest = Integer.MAX_VALUE;
		String bestcandidate = helperpositions.get(0).getAtom().getTerm(0);
		
		for (ELPLiteral e: helperpositions){
		  int distance= calcDistance(e,x,y);
		  
		  if (distance< nearest & !e.getAtom().getTerm(0).equals(name) ){ 
			  nearest = distance;
              bestcandidate = e.getAtom().getTerm(0);		   
		  }
		}
		
		return bestcandidate;
		}
		else return	cowbot.getEpistemicState().getBeliefSet().getBeliefsByPredicate("helper").get(0).getAtom().getTerm(2);
				
	}
}
