package edu.udo.cs.ie.cowbots.logic;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;

/**
 * This class models a K-program which is needed as an input for the DLV-Component for planning. Such a program consists of fluent and action - declarations, rules, an initial state of the world a goal state which has to be reached 
 * @author  Elly
 */


public class KProgram {
	
	
	public void setSecurePlan(boolean securePlan) {
		this.securePlan = securePlan;
	}
	
	public void setNoConcurrency(boolean noConcurrency) {
		this.noConcurrency = noConcurrency;
	}

	ArrayList<KDeclaration> fluents = new ArrayList<KDeclaration>();
	ArrayList<KDeclaration> actions = new ArrayList<KDeclaration>();
	ArrayList<KRule> alwaysRules = new ArrayList<KRule>();
	ArrayList<KRule> initiallyRules = new ArrayList<KRule>();
	ArrayList<ELPLiteral> goal = new ArrayList<ELPLiteral>();
	int stepsToGoal;

	ELP knowledge = new ELP();
	boolean securePlan = true;
	boolean noConcurrency = true;
	

	
	/**
	 * @param stepsToGoal  the stepsToGoal to set
	 */
	public void setStepsToGoal(int stepsToGoal) {
		this.stepsToGoal = stepsToGoal;
	}
	
	public boolean addAction(KDeclaration a) {
		return actions.add(a);
	}

	public boolean addFluent(KDeclaration f) {
		return fluents.add(f);
	}
	
	public boolean addAlwaysRule(KRule r){
		return alwaysRules.add(r);
	}
	
	public boolean addInitiallyRule(KRule r){
		return initiallyRules.add(r);
	}
	
	public boolean addGoal(ELPLiteral g) {
		return goal.add(g);
	}

	public void setGoal(ELPLiteral g){
		goal = new ArrayList<ELPLiteral>();
		goal.add(g);
	}
	
	public void addKnowledge(ELPRule k){
		knowledge.addRule(k);
	}
	/**
	 * @param elp
	 */
	public void setKnowledge (ELP elp){
		knowledge = elp;
	}


	/**
	 * @return  the knowledge
	 */
	public ELP getKnowledge() {
		return knowledge;
	}

	/**
	 * store KProgram on disc parted in two files, so it can be used by the K-solver
	 * 
	 * @param planName	output filename for the .plan part of the program
	 * @param dlName 	output filename for the .dl part of the program 
	 */
	public void saveAs(String planName, String dlName) {
		try {
			BufferedWriter w = new BufferedWriter(new FileWriter(planName+".plan"));
			w.write(planPartToString());
			w.flush();
			w.close();
			w = new BufferedWriter(new FileWriter(dlName+".dl"));
			w.write(knowledge.toString());
			w.flush();
			w.close();
		
		} catch (Exception e) {
			System.err.println(e);
		}
	}
	
	public void testsave(String name, String toSave){
		try {
			BufferedWriter w = new BufferedWriter(new FileWriter(name+".plan"));
			w.write(toSave);
			w.flush();
			w.close();
		
		} catch (Exception e) {
			System.err.println(e);
		}
	}
	
	/**
	 * store the knowledge of the program, so it can be used by the K-solver
	 * 
	 * @param dlName 	output filename for the .dl part of the program 
	 */
	public void saveKnowledge(String dlName) {
		try {
			BufferedWriter w = new BufferedWriter(new FileWriter(dlName+".dl"));
			w.write(knowledge.toString());
			w.flush();
			w.close();
		
		} catch (Exception e) {
			System.err.println(e);
		}
	}
	
	/**
	 * 
	 * @return a Sting which contains all the information of the program except the background knowledge
	 */
	
	public String planPartToString(){
		StringBuilder sb = new StringBuilder();
		
		sb.append("fluents: ");
		for(KDeclaration f : fluents) {
			sb.append(f.toString()+"\n");
		}
		sb.append("actions: ");
		for(KDeclaration a : actions) {
			sb.append(a.toString()+"\n");
		}
		sb.append("always: ");
		for(KRule r : alwaysRules) {
			sb.append(r.toString()+"\n");
		}
		sb.append("initially: ");
		for(KRule i : initiallyRules) {
			sb.append(i.toString()+"\n");
		}
		if (noConcurrency) sb.append("noConcurrency. \n" );
		if (securePlan) sb.append("securePlan. \n");
		sb.append("goal: ");
		for(ELPLiteral g : goal) {
			sb.append(g.toString());
		}
		sb.append(" ? (" + stepsToGoal + ")\n");
		return sb.toString();
		
	}
	
	public String toString (){
		String ret = "";
		ret += knowledge.toString();
		ret += planPartToString();
		return ret;
	}
	
	
	
	
	
}
