package edu.udo.cs.ie.cowbots.planer;

import jason.RevisionFailedException;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.Literal;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import edu.udo.cs.ie.cowbots.agent.ICowbot;
import edu.udo.cs.ie.cowbots.internalActions.scout.astar;
import edu.udo.cs.ie.cowbots.internalActions.scout.createWaypoints;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.logic.solver.*;

/**
 * generates plans by invoking DLV, A* or createWaypoints
 * 
 * @author Elly
 * 
 */
public class PlanGenerator {
	
	DLV dlv;

	String myself = "myself";
	
	int xAct, yAct, xGoal, yGoal;
	
	/**
	 * Generates a KProgram consisting of the knowledge of the agent and a given
	 * goal
	 * 
	 * @param ts
	 * @param goal Position the agents want to be 
	 * @param task The task the agents want to fulfill: scouted, drivenCow etc.
	 * @param args
	 * @return
	 */
	public String generate(TransitionSystem ts, ELPLiteral goal, String[] task) {
		
		myself = ts.getUserAgArch().getAgName();
		//System.out.println(myself + " Entering PlanGen: generate ");
		
		String ret = "";
		ICowbot cb = (ICowbot) ts.getAg();
		myself = ts.getUserAgArch().getAgName();
		ELP elp = cb.getEpistemicState().getBeliefsAsELP();
		ELPLiteral start = computeStartLocation(elp, task[1]);

		//System.out.println("Task1 " + task[0] + " Task2 "+ task[1]);
		
		xAct = Integer.valueOf(start.getAtom().getTerm(1));
		yAct = Integer.valueOf(start.getAtom().getTerm(2));
		xGoal = Integer.valueOf(goal.getAtom().getTerm(1));
		yGoal = Integer.valueOf(goal.getAtom().getTerm(2));
		
		// Handling if formation may cause errors when agents want to walk from the field
		// TODO Handling for the other side of the field
		if (xGoal < 1){
			xGoal=1;
			goal.getAtom().setTerm(1,Integer.toString(xGoal));
		}
		if (yGoal < 1){
			yGoal=1;
			goal.getAtom().setTerm(2,Integer.toString(yGoal));
		}
		
		System.out.println("Start: "+ start.toString() + "Ziel: " + goal.toString());
		
		// number of steps taken before planning again
		int succsteps= getSuccsteps(elp);
		System.out.println(myself + " Planning for succsteps:" + succsteps);
		
		// Aktuell wird K hier nicht genutzt.
		
		//goal = new ELPAtom("is_on", Integer.toString(xGoal), Integer.toString(yGoal));
		//KProgram program = initHigherPlanning(elp, (ELPAtom) goal);
		//String key = dlv.computePlans(program, ts);
		//System.out.println("Schlssel: " +key);

		// K entscheidet, wie geplant werden soll.
		
		astar pathfinder = new astar();	
		
		// waypoints should be created
		if (task[0].equals("scouted") || task[0].equals("driving")){
			List<Literal> wps = null;
			createWaypoints cW = new createWaypoints();
			try {
				//before creating new Waypoints the old ones have to be deleted
				deleteOldWPs(ts);
						
				//now create new Waypoints
					wps = cW.generateWaypoints(ts, xAct, yAct, xGoal, yGoal, task[0].equals("driving"));
					for(Literal wP : wps){
						ts.getAg().addBel( wP );
					}
					ret = "NumberOfWaypoints;" + wps.size();
												
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} 
			// drivenCow hier nur noch als bergangslsung drin
			// plan is created that containes walking-actions 
		} else if(task[0].startsWith("reach") || task[0].equals("drivenCow")){
			try {
				ret+= "+!"+task[0]+" : true <-";
				//ret+= ".print(Me)";
				ret += "?iam(Me);";
				String[] plan = pathfinder.getPath(ts, xAct, yAct, xGoal, yGoal, false);
				if (plan != null){
					for (int i = 0; i < Math.min(succsteps,plan.length); i++ ){
						String s = plan[i];
						ret += "!careForSwitch(" + s + ");";
//						this.computeNextPosition(s);
//						if(task[0].equals("reach") || i<plan.length-1 ){
//							ret+= "?ison(Me,"+ xAct +"," +yAct + ",_);";
//							ret+= "-failed(_);";
//						}
					}
					System.out.println("Pfadlnge: " + plan.length);
				}
							
			} catch (Exception e) {
				System.err.println("Failure in PlanGenerator");
				e.printStackTrace();
			}
			//replace last ";" by "."
			int last = ret.length()-1;
			ret = ret.substring(0,last) + ".";
			System.out.println(myself + " Plan with final dot: " + ret);
			
		} 
		
//		KProgram program = new KProgram();
//
//		program.setKnowledge(elp);
//		program = initMove(elp, program);
//		// program.addGoal(goal);
//
//
//		//ret = generateSingleMovePlan(program, start, goal);
//		ret = generateBigMovePlan(program,start,goal);
			
		System.out.println(myself +" Der Plan: " + ret);
		return ret;

	}

	
private int getSuccsteps(ELP elp) {
	List<ELPLiteral> facts = elp.getFacts();
	
	for(ELPLiteral f : facts){
		if (f.getAtom().getPredicate().equals("succstepsreconsider")){
			return Integer.valueOf(f.getAtom().getTerm(1));
		}
	}
	return 0;
}

/*
 * deletes all waypoints already in the beliefs
 * 
 */
private void deleteOldWPs(TransitionSystem ts) throws RevisionFailedException {
	
	//System.out.println("ELP vor Entfernen der WPs:" + elp.getFacts().toString());
	List<Literal> delList = new ArrayList<Literal>();
	Iterator<Literal> il = ts.getAg().getBB().iterator();
	 if (il != null) {
            while (il.hasNext()) {
                    Literal l = il.next();
                    if (l.getFunctor().equals("kwaypoint")){
                    	delList.add(l);
                }
            }
	 }
	 for (Literal l : delList){
		//System.out.println("try to abolish: " + l.toString());
		 ts.getAg().abolish(l, new Unifier()); 
	 }
	 
	//elp = cb.getEpistemicState().getBeliefsAsELP();
	//System.out.println("ELP nach Entfernen der WPs:" + elp.getFacts().toString());
	//Thread.sleep(1000);
		
	}


/**
 * computes the location the agent is at
 * 
 * @param elp knowledge of the agent
 * @param who the agent/cow for which start location is computed
 * @return
 */
	private ELPLiteral computeStartLocation(ELP elp, String who) {
	List<ELPLiteral> facts = elp.getFacts();
	
			//System.out.println("Facts: "+facts);
	for(ELPLiteral f : facts){
		if (f.getAtom().getPredicate().equals("ison") && f.getAtom().getTerm(0).equals(who)){
			ELPLiteral myLoc = new ELPAtom("is_on", f.getAtom().getArity()-1);
			myLoc.getAtom().setTerm(0, who);
			myLoc.getAtom().setTerm(1, f.getAtom().getTerm(1));
			myLoc.getAtom().setTerm(2, f.getAtom().getTerm(2));
					
			return myLoc;
				}
			}
	return null;
	}
/*
 * computes the position the agents should be at after the taking a step in direction dir
 * 
 */
private void computeNextPosition(String dir) {
	//System.out.println( dir + ": Gehe von " + xAct +" "+ yAct + " nach " );
	if (dir.equals("north")){
		yAct--  ;
	}
	if (dir.equals("northeast")){
		yAct--  ;
		xAct++  ;
	}
	if (dir.equals("east")){
		xAct++ ;
	}
	if (dir.equals("southeast")){
		xAct++ ;
		yAct++ ;
	}
	if (dir.equals("south")){
		yAct++  ;
	}

	if (dir.equals("southwest")){
		xAct--  ;
		yAct++  ;
	}

	if (dir.equals("west")){
		xAct--  ;
	}
	
	if (dir.equals("northwest")){
		xAct--  ;
		yAct--  ;
	}

	if (dir.equals("skip")){
		// nothing to do
	}
	//System.out.println(xAct +" "+ yAct);
	
	}


/*
 * initializes a KProgram for deciding which action 
 * (generating WPs, plan way with AStar) is to take
 * not needed any more
 */
private KProgram initHigherPlanning(ELP elp, ELPAtom goal) {
		KProgram prog = new KProgram();
		
		ELPAtom fluent = new ELPAtom("is_on", "X", "Y");
		ArrayList<ELPAtom> types = new ArrayList<ELPAtom>();
		types.add(new ELPAtom( "#int(X), #int(Y)"));
		KDeclaration d = new KDeclaration(fluent, types);
		prog.addFluent(d);
	
		ELPAtom action = new ELPAtom("generateWaypoints");
		types = new ArrayList<ELPAtom>();
		d = new KDeclaration(action, types);
		prog.addAction(d);
		
		action = new ELPAtom("planByAStar");
		types = new ArrayList<ELPAtom>();
		d = new KDeclaration(action, types);
		prog.addAction(d);
		
		action = new ELPAtom("planByDLVK");
		types = new ArrayList<ELPAtom>();
		d = new KDeclaration(action, types);
		prog.addAction(d);
		
		KRule krule;
		krule = new KCausationRule(new ELPAtom("is_on", "X2", "Y2"));
		krule.addIf(new NotLiteral(new ELPAtom("wp")));
		krule.addAfter(new ELPAtom("generateWaypoints"));
		prog.addAlwaysRule(krule);

		krule = new KCausationRule(new ELPAtom("is_on", "X2", "Y2"));
		krule.addIf(new ELPAtom("wp"));
		krule.addAfter(new ELPAtom("planByAStar"));
		prog.addAlwaysRule(krule);

		krule = new KExCondition(new ELPAtom("planByAStar"));
		prog.addAlwaysRule(krule);

		krule = new KExCondition(new ELPAtom("generateWaypoints"));
		prog.addAlwaysRule(krule);

		krule = new KCausationRule(new NegLiteral( new ELPAtom(goal)));
		prog.addInitiallyRule(krule);
		
		prog.addGoal(goal);
		prog.setStepsToGoal(1);
		
		prog.setKnowledge(elp);
		
		ELPRule rule;
		rule = new ELPRule();
		rule.addHead(new ELPAtom("true"));
		prog.addKnowledge(rule);

		rule = new ELPRule();
		rule.addHead(new ELPAtom("wp"));
		rule.addBody(new ELPAtom("kstartwp(_,_,_,_,_,_)"));
		prog.addKnowledge(rule);

		rule = new ELPRule();
		rule.addHead(new ELPAtom("wp"));
		rule.addBody(new ELPAtom("kendwp(_,_,_,_,_,_)"));
		prog.addKnowledge(rule);

		rule = new ELPRule();
		rule.addHead(new ELPAtom("wp"));
		rule.addBody(new ELPAtom("kdefaultwp(_,_,_,_,_,_)"));
		prog.addKnowledge(rule);

		rule = new ELPRule();
		rule.addHead(new ELPAtom("wp"));
		rule.addBody(new ELPAtom("kfenceopenwp(_,_,_,_,_,_)"));
		prog.addKnowledge(rule);
		

		rule = new ELPRule();
		rule.addHead(new ELPAtom("wp"));
		rule.addBody(new ELPAtom("kfenceclosewp(_,_,_,_,_,_)"));
		prog.addKnowledge(rule);
		
		
		
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("agent", "self"));
//		prog.addKnowledge(rule);
//		
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("object", "X"));
//		rule.addBody(new ELPAtom ("cow", "X"));
//		prog.addKnowledge(rule);
//		
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("object", "X"));
//		rule.addBody(new ELPAtom ("agent", "X"));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("agent", "X"));
//		rule.addBody(new ELPAtom ("iam", "X"));
//		prog.addKnowledge(rule);
		
		
		
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("cow", "cow1"));
//		prog.addKnowledge(rule);
				
		prog.saveAs("temp/"+myself+"/world", "temp/" +myself+"/world");
	//	System.out.println(prog.toString());
		return prog;
	}

//	/**
//	 * Computes the minimum number of steps necessary to reach goal from
//	 * startposition
//	 * 
//	 * @param goal
//	 *            Literal of the form "ison(x,y)"
//	 * @param start
//	 *            Literal of the form "ison(x,y)"
//	 * 
//	 * @return minimal number of steps necessary
//	 */
//	private int computeMinSteps(ELPLiteral start, ELPLiteral goal) {
//		int x1, y1, x2, y2, ret;
//		x1 = Integer.valueOf(start.getAtom().getTerm(1));
//		y1 = Integer.valueOf(start.getAtom().getTerm(2));
//		x2 = Integer.valueOf(goal.getAtom().getTerm(1));
//		y2 = Integer.valueOf(goal.getAtom().getTerm(2));
//		ret = (Math.abs((x1 - x2)) + Math.abs((y1 - y2)));// /2;
//		System.out.println("Minsteps: " + ret);
//		return ret;
//	}
//


//	/**
//	 * Generates a Plan for the agent
//	 * 
//	 * @param program
//	 * @param start
//	 * @param goal
//	 * @return
//	 */
//	private String generateSingleMovePlan(KProgram program, ELPLiteral start,
//			ELPLiteral goal) {
//		String ret = null;
//
//		System.out.println("start: " + start.toString() + "goal: "
//				+ goal.toString());
//		int minsteps = computeMinSteps(start, goal);
//		int maxsteps = 4 * minsteps;
//		int aktsteps = Math.max(minsteps, 1);
//
//		// mehrschrittiges Planen, Schrittweite 15
//		if (minsteps <= 15) { // mehrschrittiges Planen nicht unbedingt
//			// notwendig
//
//			System.out.println("Wahrscheinlich kein Zwischenziel notwendig");
//			while (aktsteps < Math.max(maxsteps, 15) && ret == null) {
//				program.setStepsToGoal(aktsteps);
//				// System.out.println(program);
//				System.out.println("Versuche mit #Schritten:" + aktsteps);
//				program.setGoal(goal);
//				KCausationRule startRule = new KCausationRule(start);
//				program.addInitiallyRule(startRule);
//
//				program.saveAs("world", "world");
//				ret = dlv.computePlans(program);
//				aktsteps++;
//			}
//			if (ret != null) {
//				System.out.println("Kein Zwischenziel notwendig");
//				return ret;
//			}
//		}
//		// mehrschrittiges Planen notwendig, erstelle zwischenziele
//		// diagonales zwischenziel
//
//		int x1, y1, x2, y2, xm, ym;
//
//		x1= Integer.valueOf(start.getAtom().getTerm(1));
//		y1= Integer.valueOf(start.getAtom().getTerm(2));
//		x2= Integer.valueOf(goal.getAtom().getTerm(1));
//		y2= Integer.valueOf(goal.getAtom().getTerm(2));
//		xm = (x1+x2)/2;
//		ym = (y1+y2)/2;
//		ELPAtom subGoal = new ELPAtom("is_on", myself,xm+"",ym+"");
//
//		System.out.println("Diagonales ZwischenZiel: " + subGoal);
//		String tmp = generateSingleMovePlan(program, start, subGoal);
//		if (tmp != null) {
//			ret = tmp;
//			// System.out.println(tmp +" Zwischenplan1 " + ret);
//		} else { // Planen fehlgeschlagen
//			System.out.println("Fehler beim Planen");
//		}
//		
//		tmp = generateSingleMovePlan(program, subGoal, goal);
//		if (tmp != null) {
//			ret += tmp;
//			// System.out.println( tmp+ " Zwischenplan2 " + ret);
//		} else {
//			System.out.println("Fehler beim Planen");
//		}
//
//		// horizontales zwischenziel
//		// vertikales zwischenziel
//
//		return ret;
//	}
//

/*
 * computes X and Y for the position the agent will be after move
 * Used to compute testgoals
 * 
 * @param dir: Direction the agent moves
 */

//private String generateBigMovePlan(KProgram program, ELPLiteral start, ELPLiteral goal) {
//		String ret = null;
//
//		System.out.println("Ziel: " + goal.toString() + "Start: " + start.toString());
//		int steps = 1;
//		while ( steps <20 && ret == null){
//		//for (int steps = 1; steps < 20; steps++){
//			program.setStepsToGoal(steps);
//			System.out.println("Versuche mit #Schritten:" + steps);
//			program.setGoal(goal);
//			//KCausationRule startRule = new KCausationRule(start);
//			//program.addInitiallyRule(startRule);
//
//			
//			program.saveAs("world", "beliefs");
//			ret = dlv.computePlans(program);
//			steps++;
//		}
//		
//		if (ret == null){
//			ret += "skip";
//		}
//		return ret;
//	}
//

//	/**
//	 * initMove adds rules and knowledge necessary to call dlv in order to find
//	 * a path step by step
//	 * 
//	 * @param program
//	 * 
//	 * @return program can be used to call dlv.computePlans
//	 */
//	// TODO: Ist momentan zu explizit, ein Parser, der die ganzen Regeln
//	// einliest wre schner.
//
//	private KProgram initMove(ELP elp, KProgram prog) {
//
//		// Fr das big_move Planen
//		
//		ELPAtom fluent = new ELPAtom("is_on", "X", "Y");
//		ArrayList<ELPAtom> types = new ArrayList<ELPAtom>();
//		types.add(new ELPAtom("#int(X), #int(Y)"));
//		KDeclaration d = new KDeclaration(fluent, types);
//		prog.addFluent(d);
//		
//		ELPAtom action = new ELPAtom("big_move","D", "X1", "Y1", "X2", "Y2");
//		types = new ArrayList<ELPAtom>();
//		types.add(new ELPAtom("#int(D), #int(X1), #int(Y1), #int(X2), #int(Y2)"));
//		d = new KDeclaration(action, types);
//		prog.addAction(d);
//		
//		KRule krule;
//		krule = new KCausationRule(new ELPAtom("is_on", "X2", "Y2"));
//		krule.addAfter(new ELPAtom("big_move", "D", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);
//
//		krule = new KCausationRule(new NegLiteral(new ELPAtom("is_on", "X1", "Y1")));
//		krule.addAfter(new ELPAtom("big_move", "D", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X2", "Y2"));
//		krule.addIf(new ELPAtom("X1 <> X2"));
//		krule.addIf(new ELPAtom("Y1 <> Y2"));
//		prog.addAlwaysRule(krule);
//		
//		// TODO inertial is_on !
//		
//		krule = new KExCondition(new ELPAtom("big_move", "1", "X", "Y1", "X", "Y2"));
//		krule.addIf(new ELPAtom("free", "X", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X", "Y1"));
//		krule.addIf(new ELPAtom("freeNLine", "X", "Y1", "Y2"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "2", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("free", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y1"));
//		krule.addIf(new ELPAtom("freeNELine", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "3", "X1", "Y", "X2", "Y"));
//		krule.addIf(new ELPAtom("free", "X2", "Y"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y"));
//		krule.addIf(new ELPAtom("freeELine", "X1", "X2", "Y"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "4", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("free", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y1"));
//		krule.addIf(new ELPAtom("freeSELine", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "5", "X", "Y1", "X", "Y1"));
//		krule.addIf(new ELPAtom("free", "X", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X", "Y1"));
//		krule.addIf(new ELPAtom("freeSLine", "X", "Y1", "Y2"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "6", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("free", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y1"));
//		krule.addIf(new ELPAtom("freeSWLine", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "7", "X1", "Y", "X2", "Y"));
//		krule.addIf(new ELPAtom("free", "X2", "Y"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y"));
//		krule.addIf(new ELPAtom("freeWLine", "X1", "X2", "Y"));
//		prog.addAlwaysRule(krule);		
//		krule = new KExCondition(new ELPAtom("big_move", "8", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("free", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "X1", "Y1"));
//		krule.addIf(new ELPAtom("freeNWLine", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);
//		
//
//		krule = new KCausationRule(new ELPAtom("is_on", "X", "Y"));
//		krule.addIf(new ELPAtom("ison", myself, "X", "Y", "_"));
//		prog.addInitiallyRule(krule);
//		
//		ELPRule rule;
//		
//		List<ELPLiteral> facts = prog.getKnowledge().getFacts();
//		int wzx = 0, wzy = 0;
//		for (ELPLiteral f : facts) {
//			if (f.getAtom().getPredicate().equals("worldsize")) {
//				wzx = Integer.valueOf(f.getAtom().getTerm(0));
//				wzy = Integer.valueOf(f.getAtom().getTerm(1));
//			}
//		}
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("free", "X", "Y"));
//		rule.addBody(new ELPAtom("X>0, Y>0,  #int(X), #int(Y), #int(C)"));
//		rule.addBody(new NotLiteral(new ELPAtom("ison", "obstacle", "X", "Y", "C")));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeNLine", "X,Y1,Y2"));	
//		rule.addBody(new ELPAtom("Y1=Y2, free(X,Y1), free(X,Y2)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeNLine", "X,Y1,Y2"));	
//		rule.addBody(new ELPAtom("H=Y1+1, freeNLine(X,H,Y2), free(X,H)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeSLine", "X,Y1,Y2"));
//		rule.addBody(new ELPAtom("freeNLine", "X,Y2,Y1"));				
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		
//		rule.addHead(new ELPAtom("freeELine", "X1,X2,Y"));	
//		rule.addBody(new ELPAtom("X1=X2, free(X1,Y), free(X2,Y)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeELine", "X1,X2,Y"));	
//		rule.addBody(new ELPAtom("V=X1+1, freeELine(V,X2,Y), free(V,Y)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeWLine", "X1,X2,Y"));
//		rule.addBody(new ELPAtom("freeELine", "X2,X1,Y"));				
//		prog.addKnowledge(rule);
//		
//		rule=new ELPRule();
//		rule.addHead(new ELPAtom("freeNELine", "X1,X2,Y1,Y2"));	
//		rule.addBody(new ELPAtom("X1=X2,Y1=Y2, free(X1,Y1), free(X2,Y2)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeNELine", "X1,X2,Y1,Y2"));	
//		rule.addBody(new ELPAtom("#succ(Y1,H), #succ(X1,V), freeNELine(V,H,X2,Y2), free(V,H)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeSWLine", "X1,X2,Y1,Y2"));
//		rule.addBody(new ELPAtom("freeNELine", "X2,Y2,X1,Y1"));				
//		prog.addKnowledge(rule);
//
//		rule= new ELPRule();
//		rule.addHead(new ELPAtom("freeSELine", "X1,X2,Y1,Y2"));	
//		rule.addBody(new ELPAtom("X1=X2,Y1=Y2, free(X1,Y1), free(X2,Y2)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("freeSELine", "X1,X2,Y1,Y2"));	
//		rule.addBody(new ELPAtom("#succ(H,Y1), #succ(X1,V), freeSELine(V,H,X2,Y2), free(V,H)"));
//		prog.addKnowledge(rule);
//		rule = new ELPRule();
//		
//		rule.addHead(new ELPAtom("freeNWLine", "X1,X2,Y1,Y2"));
//		rule.addBody(new ELPAtom("freeSELine", "X2,Y2,X1,Y1"));				
//		prog.addKnowledge(rule);
//
//		
//		
		// Ab hier fr das single_move Planen
		
//		ELPAtom fluent = new ELPAtom("is_on", "AG", "X", "Y");
//		ArrayList<ELPAtom> types = new ArrayList<ELPAtom>();
//		types.add(new ELPAtom("agent", "AG"));
//		types.add(new ELPAtom("location", "X", "Y"));
//		KDeclaration d = new KDeclaration(fluent, types);
//		prog.addFluent(d);
//
//		fluent = new ELPAtom("obstacle_on", "X", "Y");
//		types = new ArrayList<ELPAtom>();
//		types.add(new ELPAtom("location", "X", "Y"));
//		d = new KDeclaration(fluent, types);
//		prog.addFluent(d);
//
//		ELPAtom action = new ELPAtom("single_move", "AG", "X1", "Y1", "X2", "Y2");
//		types = new ArrayList<ELPAtom>();
//		types.add(new ELPAtom("agent", "AG"));
//		types.add(new ELPAtom("location", "X1", "Y1"));
//		types.add(new ELPAtom("location", "X2", "Y2"));
//		d = new KDeclaration(action, types);
//		prog.addAction(d);
//
//		KRule krule;
//		krule = new KCausationRule(new ELPAtom("is_on", "AG", "X2", "Y2"));
//		krule.addAfter(new ELPAtom("single_move", "AG", "X1", "Y1", "X2", "Y2"));
//		prog.addAlwaysRule(krule);
//
//		krule = new KCausationRule(new NegLiteral(new ELPAtom("is_on", "AG", "X1", "Y1")));
//		krule.addAfter(new ELPAtom("single_move", "AG", "X1", "Y1", "X2", "Y2"));
//		krule.addAfter(new ELPAtom("is_on", "AG", "X2", "Y2"));
//		krule.addAfter(new ELPAtom("X1 <> X2"));
//		krule.addAfter(new ELPAtom("Y1 <> Y2"));
//		prog.addAlwaysRule(krule);
//
//		krule = new KExCondition(new ELPAtom("single_move", "AG", "X1", "Y1",
//				"X2", "Y2"));
//		krule.addIf(new ELPAtom("neigh", "X1", "Y1", "X2", "Y2"));
//		krule.addIf(new ELPAtom("is_on", "AG", "X1", "Y1"));
//		krule.addIf(new NotLiteral(new ELPAtom("obstacle_on", "X2", "Y2")));
//		prog.addAlwaysRule(krule);
//
//		krule = new KCausationRule(new ELPAtom("obstacle_on", "X", "Y"));
//		krule.addIf(new ELPAtom("ison", "obstacle", "X", "Y", "_"));
//		krule.addIf(new ELPAtom("obstacle", "obstacle"));
//		prog.addAlwaysRule(krule);
//
//		krule = new KCausationRule(new ELPAtom("is_on", "myself", "X", "Y"));
//		// krule.addIf(new ELPAtom ("iam", "AG"));
//		krule.addIf(new ELPAtom("ison", "myself", "X", "Y", "_"));
//		krule.addIf(new NotLiteral(new ELPAtom("is_on", "myself", "A", "B")));
//		krule.addIf(new ELPAtom("A<>X, B<>Y"));
//		prog.addInitiallyRule(krule);
//
//		ELPRule rule;
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("neigh", "X1", "Y1", "X2", "Y2"));
//		rule.addBody(new ELPAtom("location(X1,Y1)"));
//		rule.addBody(new ELPAtom("location(X2,Y2)"));
//		rule.addBody(new ELPAtom("X1=X2"));
//		rule.addBody(new ELPAtom("Y1=Y2+1"));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("neigh", "X1", "Y1", "X2", "Y2"));
//		rule.addBody(new ELPAtom("location(X1,Y1)"));
//		rule.addBody(new ELPAtom("location(X2,Y2)"));
//		rule.addBody(new ELPAtom("X1=X2+1"));
//		rule.addBody(new ELPAtom("Y1=Y2"));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("neigh", "X1", "Y1", "X2", "Y2"));
//		rule.addBody(new ELPAtom("location(X1,Y1)"));
//		rule.addBody(new ELPAtom("location(X2,Y2)"));
//		rule.addBody(new ELPAtom("X1=X2+1"));
//		rule.addBody(new ELPAtom("Y1=Y2+1"));
//		prog.addKnowledge(rule);
//
//		List<ELPLiteral> facts = prog.getKnowledge().getFacts();
//		int wzx = 0, wzy = 0;
//		for (ELPLiteral f : facts) {
//			if (f.getAtom().getPredicate().equals("worldsize")) {
//				wzx = Integer.valueOf(f.getAtom().getTerm(0));
//				wzy = Integer.valueOf(f.getAtom().getTerm(1));
//			}
//		}
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("location", "X", "Y"));
//		rule.addBody(new ELPAtom("#int(X), #int(Y), X<=" + wzx + ", Y<="+ wzy));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("neigh", "X1", "Y1", "X2", "Y2"));
//		rule.addBody(new ELPAtom("neigh", "X2", "Y2", "X1", "Y1"));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("agent", "AG"));
//		rule.addBody(new ELPAtom("ally", "AG"));
//		prog.addKnowledge(rule);
//
//		rule = new ELPRule();
//		rule.addHead(new ELPAtom("agent", "myself"));
//		prog.addKnowledge(rule);

//		return prog;
//	}

}
