package edu.udo.cs.ie.cowbots.util;

import java.awt.Point;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.*;

import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.bdi.brf.IdentTable;

/**
 * basic class to model a two dimensional map.
 * 
 * class is preliminary and subject to change!
 * 
 * @author Thomas
 *
 * @param <T>	content type
 */
public class GridMap {
	
	ArrayList< ArrayList<GridCell<ELPLiteral>>>	gridmap;
	int w,h;
	
	private class Coordinate {
		public Coordinate(int x, int y) {
			this.x = x;
			this.y = y;
		}
		
		public int x;
		public int y;
	}
	
	List<Coordinate> movers;	
	IdentTable idents;
	/*
	List<Coordinate> switches;
	List<Point> fences; 
	
	*/
	
	public GridMap() {
		gridmap = null;
		w = 0;
		h = 0;
		movers = new LinkedList<Coordinate>();
		
		idents = null;
		
	}
	
	public void setIdentTable(IdentTable it) {
		this.idents = it;
	}
	
	public void setDimension(int x, int y) {
		gridmap = new ArrayList< ArrayList<GridCell<ELPLiteral>> >(y);
		for (int i = 0; i < y; i++) {
			ArrayList<GridCell<ELPLiteral>> yy = new ArrayList<GridCell<ELPLiteral>>();
			gridmap.add( yy );
			for (int j = 0; j < x; j++) {
				yy.add(new GridCell<ELPLiteral>());
			}
		}
		w = x;
		h = y;
	}
	
	public void setDimensionContest(int x, int y, int minCX, int minCY, int maxCX, int maxCY) {		
		gridmap = new ArrayList< ArrayList<GridCell<ELPLiteral>> >(y);
		for (int i = 0; i < y; i++) {
			ArrayList<GridCell<ELPLiteral>> yy = new ArrayList<GridCell<ELPLiteral>>();
			gridmap.add( yy );
			for (int j = 0; j < x; j++) {
				GridCell<ELPLiteral> cell = new GridCell<ELPLiteral>();				
				if ((j >= minCX) && (j<= maxCX) && (i >= minCY) && (i <= maxCY)) {
					cell.cell_content = new ELPAtom("ison","owncorral",""+j,""+i,""+0);
					cell.isUnexplored = true;
				} else {
					cell.cell_content = new ELPAtom("ison","unexplored",""+j,""+i,""+0);
					cell.isUnexplored = false;
				}
				yy.add(cell);
			}
		}
		w = x;
		h = y;
		
		movers.clear();	
	}
	
	public void setFlag(int x, int y, ELPLiteral data) {
		if (( x >= this.w ) || ( y >= this.h ))
			return;
		GridCell<ELPLiteral> cell = gridmap.get(y).get(x);
		cell.cell_flag = data;
		
		// check for isopen flag
		if (data.getAtom().getTerm(2).equalsIgnoreCase("true")) {
			cell.isClosed = false;
		} else
			cell.isClosed = true;
		
	}
		
	
	/**
	 * this method updates a fence state depending the given
	 * data has a greater time value than the present one.
	 * 
	 * @param x
	 * @param y
	 * @param data
	 * @param timeTerm
	 */
	public void setFlagNewer(int x, int y, ELPLiteral data, int timeTerm) {
		if (( x >= this.w ) || ( y >= this.h ))
			return;
		
		GridCell<ELPLiteral> cell = gridmap.get(y).get(x);
		ELPLiteral old = cell.cell_flag;
		if (old == null) {
			cell.cell_flag = data;
		} else {
			int t_old = old.getAtom().getTermAsInt(timeTerm);
			int t_new = data.getAtom().getTermAsInt(timeTerm);
			if (t_new > t_old)
				cell.cell_flag = data;
		}
		
		cell.isFence = true;
		// check for isopen flag
		if (cell.cell_flag.getAtom().getTerm(2).equalsIgnoreCase("true")) {
			cell.isClosed = false;
		} else
			cell.isClosed = true;
	}
	
	public ELPLiteral setContent(int x, int y, ELPLiteral data) {
		if (( x >= this.w ) || ( y >= this.h ))
			return null;
		
		ELPLiteral ret = null;
		GridCell<ELPLiteral> cell = this.getCell(x, y);
		
		if (this.idents.isKindOf(data.getAtom().getTerm(0), "obstacle"))
			cell.isObstacle = true;
		else if (this.idents.isKindOf(data.getAtom().getTerm(0), "switch"))
			cell.isSwitch = true;
		else if (this.idents.isKindOf(data.getAtom().getTerm(0), "fence"))
			cell.isFence = true;
		
		cell.isUnexplored = false;
		
		if (cell.has_mover) {
			ret = cell.cell_type;
			cell.cell_type = data;
		} else {
			ret = cell.cell_content;
			cell.cell_content = data;
		}
		return ret;
	}
	
	/**
	 * returns the "ison" predicate at a given
	 * (x,y) coordinate index
	 * 
	 * @param x		x-coordinate
	 * @param y		y-coordinate
	 * @return
	 */
	public ELPLiteral getContent(int x, int y) {
		return gridmap.get(y).get(x).cell_content;
	}
	
	
	/**
	 * returns the type of a cell, or a null pointer.
	 * returns usefull information only if getContent
	 * returned a cow or an agent
	 * @param x	cell coordinate
	 * @param y cell coordinate
	 * @return	null or an ELPLiteral describing a cell type (fence, corral)
	 */
	public ELPLiteral getBgnd(int x, int y) {
		return gridmap.get(y).get(x).cell_type;
	}
		
	
	/**
	 * returns the "isopen" literal at a given
	 * location (x,y)
	 * @param x		x-coordinate
	 * @param y		y-coordinate
	 * @return	null or isopen-literal
	 */
	public ELPLiteral getFlag(int x, int y) {
		return gridmap.get(y).get(x).cell_flag;
	}
	
		
	/**
	 * returns the width, in tiles, of this map
	 * @return	width of map
	 */
	public int width() {
		return w;
	}
	
	/**
	 * returns the height, in tiles, of this map
	 * @return	height of map
	 */
	public int height() {
		return h;
	}

	/**
	 * resets all moving entities (removed from map)
	 */
	public void resetMovers() {
		// clear agent positions from map
		for (Coordinate coord : movers) {
			GridCell<ELPLiteral> cell = getCell(coord.x, coord.y);
			cell.cell_content = cell.cell_type;
			cell.has_mover = false;
			cell.isAgent = false;
			cell.isCow = false;
		}
		movers.clear();
	}

	/**
	 * updates cell content with moving entities
	 * @param movingObjects		collection of moving entities
	 */
	public void updateMovers(Collection<ELPLiteral> movingObjects) {		
		for (ELPLiteral l : movingObjects) {
			String movertype = l.getAtom().getTerm(0);
			int x = Integer.parseInt(l.getAtom().getTerm(1));
			int y = Integer.parseInt(l.getAtom().getTerm(2));
			Coordinate coord = new Coordinate(x,y);
			movers.add(coord);
			GridCell<ELPLiteral> cell = getCell(x,y);
			cell.isUnexplored = false;
			if (!cell.has_mover) {
				cell.cell_type = cell.cell_content;
				cell.cell_content = l;
				cell.has_mover=true;
				if (idents.isKindOf(movertype, "agent"))
					cell.isAgent = true;
				if (idents.isKindOf(movertype, "cow"))
					cell.isCow = true;
			} else {
				cell.cell_content = l;
			}
		}
	}
	
	
	public GridCell<ELPLiteral> getCell(int x, int y) {
		return this.gridmap.get(y).get(x);
	}
	
	public void saveAs(String filename) {
		try {
			BufferedWriter w = new BufferedWriter(new FileWriter(filename));
			for (int y = 0; y < this.h; y++) {
				String line = "";
				for(int x = 0; x < this.w; x++) {
					GridCell<ELPLiteral> cell = getCell(x,y);
					char content = '.';
					ELPLiteral l = cell.cell_content;
					if (l == null) {
						line += content;
						continue;
					}
					String type = l.getAtom().getTerm(0);
					if (idents.isKindOf(type, "obstacle"))
						content = '#';
					else if (idents.isKindOf(type, "fence"))
						content = 'F';
					else if (idents.isKindOf(type, "switch"))
						content = 'S';
					else if (idents.isKindOf(type, "iam"))
						content = 'I';
					else if (idents.isKindOf(type, "ally"))
						content = 'A';
					else if (idents.isKindOf(type, "enemy"))
						content = 'E';
					else if (idents.isKindOf(type, "cow"))
						content = 'C';
					else if (idents.isKindOf(type, "owncorral"))
						content = 'a';
					else if (idents.isKindOf(type, "enemycorral"))
						content = 'e';
					else if (idents.isKindOf(type, "unexplored"))
						content = 'u';
					line += content;
					
				}
					
				w.write(line);
				w.newLine();
			}
			w.flush();
			w.close();
		} catch (Exception e) {
			
		}
	}
	
	
	/**
	 * returns a list of all moving objects by a given type
	 * 
	 * @param identPredicate
	 * @return
	 */
	public List<ELPLiteral> getMoversByType(String identPredicate) {
		LinkedList<ELPLiteral> ret = new LinkedList<ELPLiteral>();
		for (Coordinate c:movers) {
			ELPLiteral l = this.getContent(c.x, c.y);
			if ( this.idents.isKindOf( l.getAtom().getTerm(0), identPredicate) )
				ret.add(l);
		}		
		return ret;
	}
	
	/**
	 * returns the distance map associated with this grid map, and
	 * all obstacles are put on the map initially. you may optionally
	 * enable other things as obstacles.
	 * @return
	 */
	public DistanceMap getDistanceMap(boolean agentsAsObstacles) {
		
		
		return getDistanceMap(agentsAsObstacles, false);
	}
	
	/**
	 * returns a distance map, but closed fences are treated as obstacles
	 * (if 2nd argument set to true).
	 * @param agentsAsObstacles
	 * @param closedFences
	 * @return
	 */
	public DistanceMap getDistanceMap(boolean agentsAsObstacles, boolean closedFences) {
		DistanceMap dm = new DistanceMap( this.w, this.h );
		for (int j = 0; j < this.h; j++)
			for (int i = 0; i < this.w; i++) {
				GridCell<ELPLiteral> gc = this.gridmap.get(j).get(i);
				
				// set various bit flags
				if (gc.isObstacle)
					dm.setObstacle(i,j);
				if (gc.isSwitch) {
					dm.setObstacle(i,j);
					dm.setAtSwitch(i+1, j);
					dm.setAtSwitch(i-1, j);
					dm.setAtSwitch(i, j+1);
					dm.setAtSwitch(i, j-1);
				}
				
				if (gc.isClosed && closedFences) {
					dm.setObstacle(i, j);
					dm.setPenalty(i, j, 1);
				} else if (gc.isClosed) {
					dm.setPenalty(i, j, 1);
				}
								
			}
		
		if (agentsAsObstacles) {
			for (Coordinate coord : movers) {
				GridCell<ELPLiteral> gc = this.gridmap.get(coord.y).get(coord.x);
				if (gc.isAgent)
					dm.setObstacle(coord.x, coord.y);
			}
		}
		return dm;
	}
		
	public boolean isFence(int x, int y) {
		if ((x < 0) || (y < 0) || (x >= this.w) || (y >= this.h))
			return false;
		
		return this.gridmap.get(y).get(x).isFence;
	}
	
	
	/**
	 * returns true if a cell has at least two adjacent obstacle
	 * or switch cells. used in the contest to prevent clustering
	 * cows that are in a corner.
	 * @param x
	 * @param y
	 * @return
	 */
	public boolean isCorner(int x, int y) {
		
		GridCell<ELPLiteral> cell = null;
		int corners = 0;
		
		// qickfix: interpret agents as obstacles too
		// otherwise, they stall on cows to quickly
		
		int dX = x+1;
		int dY = y;
		if (dX < this.w) {
			cell = this.getCell(dX, dY);
			// if (cell.isObstacle || cell.isSwitch)
			if (cell.isObstacle || cell.isSwitch || cell.isAgent)
				++corners;
		} else
			++corners;
		
		dX = x-1;
		if (dX >= 0) {
			cell = this.getCell(dX, dY);
			if (cell.isObstacle || cell.isSwitch || cell.isAgent)
				++corners;
		} else
			++corners;
		
		dX = x;
		dY = y-1;
		if (dY >= 0) {
			cell = this.getCell(dX, dY);
			if (cell.isObstacle || cell.isSwitch || cell.isAgent)
				++corners;
		} else
			++corners;
		
		dY = y+1;
		if (dY < this.h) {
			cell = this.getCell(dX, dY);
			if (cell.isObstacle || cell.isSwitch || cell.isAgent)
				++corners;
		} else
			++corners;
		
		return (corners >= 2)?true : false;
	}
	
	
	public int getFenceOpenTime(int x, int y) {
		ELPLiteral l = this.getFlag(x, y);
		if (l == null)
			return 0;
		return l.getAtom().getTermAsInt(3);
	}

	public void setGateState(List<Point> fenceCells, int fenceTimeMax, String fenceState) {
		for (Point p : fenceCells) {
			GridCell<ELPLiteral> c = this.gridmap.get(p.y).get(p.x);
			c.cell_flag.getAtom().setTerm(2, fenceState);
			c.cell_flag.getAtom().setTerm(3, ""+fenceTimeMax);
		}
	}
}
